from __future__ import annotations

import warnings
import weakref
from copy import copy
from functools import wraps
from inspect import getcallargs
from typing import Any, Dict, List, Mapping, Sequence, Tuple, Union

import meerkat as mk
from meerkat.errors import ExperimentalWarning

_provenance_enabled = False


def set_provenance(enabled=True):
    global _provenance_enabled
    _provenance_enabled = enabled


class provenance:
    def __init__(self, enabled: bool = True):
        """Context manager for enabling provenance capture in Meerkat.

        Example:
        ```python
        import meerkat as mk
        with mk.provenance():
            dp = mk.DataPanel.from_batch({...})
        ```
        """
        self.prev = None
        self.enabled = enabled

    def __enter__(self):
        global _provenance_enabled
        self.prev = _provenance_enabled
        _provenance_enabled = self.enabled

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any):
        global _provenance_enabled
        _provenance_enabled = self.prev


def is_provenance_enabled():
    return _provenance_enabled


class ProvenanceMixin:
    def __init__(self, *args, **kwargs):
        super(ProvenanceMixin, self).__init__(*args, **kwargs)
        self._init_node()

    def _init_node(self):
        self._node = ProvenanceObjNode(self)

    @property
    def node(self):
        return self._node

    def get_provenance(
        self, include_columns: bool = False, last_parent_only: bool = False
    ):
        if self._node is None:
            return None
        nodes, edges = self._node.get_provenance(last_parent_only=last_parent_only)
        if include_columns:
            return nodes, edges
        else:
            # filter out edges and nodes
            nodes = [
                node
                for node in nodes
                if (
                    (
                        isinstance(node, ProvenanceOpNode)
                        and any(
                            [
                                issubclass(node.type, mk.DataPanel)
                                for node, key in node.children
                            ]
                        )
                    )
                    or issubclass(node.type, mk.DataPanel)
                )
            ]
            edges = [
                edge for edge in edges if all([(node in nodes) for node in edge[:2]])
            ]
            return nodes, edges


class ProvenanceNode:
    def add_child(self, node: ProvenanceNode, key: Tuple):
        self._children.append((node, key))
        with provenance(enabled=False):
            self.cache_repr()

    def add_parent(self, node: ProvenanceNode, key: Tuple):
        self._parents.append((node, key))
        with provenance(enabled=False):
            self.cache_repr()

    @property
    def parents(self):
        return self._parents

    @property
    def last_parent(self):
        if len(self._parents) > 0:
            return self._parents[-1]
        else:
            return None

    @property
    def children(self):
        return self._children

    def get_provenance(self, last_parent_only: bool = False):
        edges = set()
        nodes = set()
        for parent, key in (
            self.parents[-1:]
            if (last_parent_only and self.__class__ == ProvenanceObjNode)
            else self.parents
        ):
            prev_nodes, prev_edges = parent.get_provenance(
                last_parent_only=last_parent_only
            )
            edges |= prev_edges
            edges.add((parent, self, key))
            nodes |= prev_nodes

        nodes.add(self)
        return nodes, edges

    def cache_repr(self):
        self._cached_repr = self._repr() + "(deleted)"

    def _repr(self):
        raise NotImplementedError()

    def __repr__(self):
        if self.ref() is None:
            return self._cached_repr
        else:
            return self._repr()

    def __getstate__(self):
        state = copy(self.__dict__)
        state["ref"] = None
        return state

    def __setstate__(self, state: dict):
        state["ref"] = lambda: None
        self.__dict__.update(state)


class ProvenanceObjNode(ProvenanceNode):
    def __init__(self, obj: ProvenanceMixin):
        self.ref = weakref.ref(obj)
        self.type = type(obj)
        self._parents: List[Tuple(ProvenanceOpNode, tuple)] = []
        self._children: List[Tuple(ProvenanceOpNode, tuple)] = []

    def _repr(self):
        return f"ObjNode({str(self.ref())})"


class ProvenanceOpNode(ProvenanceNode):
    def __init__(
        self, fn: callable, inputs: dict, outputs: object, captured_args: dict
    ):
        self.ref = weakref.ref(fn)
        self.type = type(fn)
        self.name = fn.__qualname__
        self.module = fn.__module__
        self.captured_args = captured_args

        self._parents: List[Tuple(ProvenanceObjNode, tuple)] = []
        self._children: List[Tuple(ProvenanceObjNode, tuple)] = []

        for key, inp in inputs.items():
            self.add_parent(inp.node, key)
            inp.node.add_child(self, key)

        input_ids = {id(v) for v in inputs.values()}
        for key, output in outputs.items():
            # only include outputs that weren't passed in
            if id(output) not in input_ids:
                self.add_child(output.node, key)
                output.node.add_parent(self, key)

    def _repr(self):
        return f"OpNode({self.module}.{self.name})"


def capture_provenance(capture_args: Sequence[str] = None):
    capture_args = [] if capture_args is None else capture_args

    def _provenance(fn: callable):
        @wraps(fn)
        def _wrapper(*args, **kwargs):
            if not is_provenance_enabled():
                return fn(*args, **kwargs)
            args_dict = getcallargs(fn, *args, **kwargs)
            if "kwargs" in args_dict:
                args_dict.update(args_dict.pop("kwargs"))

            captured_args = {arg: args_dict[arg] for arg in capture_args}
            out = fn(*args, **kwargs)

            # collect instances of ProvenanceMixin nested in the output
            inputs = get_nested_objs(args_dict)
            outputs = get_nested_objs(out)
            ProvenanceOpNode(
                fn=fn, inputs=inputs, outputs=outputs, captured_args=captured_args
            )
            return out

        return _wrapper

    return _provenance


def get_nested_objs(data):
    """Recursively get DataPanels and Columns from nested collections."""
    objs = {}
    _get_nested_objs(objs, (), data)
    return objs


def _get_nested_objs(objs: Dict, key: Tuple[str], data: object):
    if isinstance(data, Sequence) and not isinstance(data, str):
        for idx, item in enumerate(data):
            _get_nested_objs(objs, key=(*key, idx), data=item)

    elif isinstance(data, Mapping):
        for curr_key, item in data.items():
            _get_nested_objs(objs, key=(*key, curr_key), data=item)
    elif isinstance(data, mk.DataPanel):
        objs[key] = data
        for curr_key, item in data.items():
            _get_nested_objs(objs, key=(*key, curr_key), data=item)
    elif isinstance(data, mk.AbstractColumn):
        objs[key] = data


def visualize_provenance(
    obj: Union[ProvenanceObjNode, ProvenanceOpNode],
    show_columns: bool = False,
    last_parent_only: bool = False,
):

    warnings.warn(  # pragma: no cover
        ExperimentalWarning(
            "The function `meerkat.provenance.visualize_provenance` is experimental and"
            " has limited test coverage. Proceed with caution."
        )
    )
    try:  # pragma: no cover
        import cyjupyter
    except ImportError:  # pragma: no cover
        raise ImportError(
            "`visualize_provenance` requires the `cyjupyter` dependency."
            "See https://github.com/cytoscape/cytoscape-jupyter-widget"
        )

    nodes, edges = obj.get_provenance(  # pragma: no cover
        include_columns=show_columns, last_parent_only=last_parent_only
    )
    cy_nodes = [  # pragma: no cover
        {
            "data": {
                "id": id(node),
                "name": str(node),
                "type": "obj" if isinstance(node, ProvenanceObjNode) else "op",
            }
        }
        for node in nodes
    ]
    cy_edges = [  # pragma: no cover
        {
            "data": {
                "source": id(edge[0]),
                "target": id(edge[1]),
                "key": edge[2],
            }
        }
        for edge in edges
    ]

    cy_data = {"elements": {"nodes": cy_nodes, "edges": cy_edges}}  # pragma: no cover

    style = [  # pragma: no cover
        {
            "selector": "node",
            "css": {
                "content": "data(name)",
                "background-color": "#fc8d62",
                "border-color": "#252525",
                "border-opacity": 1.0,
                "border-width": 3,
            },
        },
        {
            "selector": "node[type = 'op']",
            "css": {
                "shape": "barrel",
                "background-color": "#8da0cb",
            },
        },
        {
            # need to double index to access metadata (degree etc.)
            "selector": "node[[degree = 0]]",
            "css": {
                "visibility": "hidden",
            },
        },
        {
            "selector": "edge",
            "css": {
                "content": "data(key)",
                "line-color": "#252525",
                "mid-target-arrow-color": "#8da0cb",
                "mid-target-arrow-shape": "triangle",
                "arrow-scale": 2.5,
                "text-margin-x": 10,
                "text-margin-y": 10,
            },
        },
    ]
    return cyjupyter.Cytoscape(  # pragma: no cover
        data=cy_data, visual_style=style, layout_name="breadthfirst"
    )



import random
from typing import Optional, Tuple

import torch


class TemporalDownsampling(object):
    """Video transformation for performing temporal downsampling (i.e. reading
    in every Nth frame only). This can be used in tandem with VideoCell by
    passing it into the `transform` keyword in the constructor. Can be used
    with Compose in torchvision.

    When using with TemporalCrop, it is highly recommended to put TemporalDownsampling
    first, with TemporalCrop second.

    Arguments:
        downsample_factor (int): the factor by which the input video should be
            downsampled. Must be a strictly positive integer.
        time_dim (int): the time dimension of the input video.

    Examples:

        # Create a VideoCell from "/path/to/video.mp4" with time in dimension one,
        showing every other frame
        >>> cell = VideoCell("path/to/video.mp4",
            time_dim=1,
            transform=TemporalDownsampling(2, time_dim=1)
            )

    Note that time_dim in the TemporalDownsampling call must match the the time_dim
    in the VideoCell constructor!
    """

    def __init__(self, downsample_factor: int, time_dim: Optional[int] = 1):
        self.downsample_factor = downsample_factor
        self.time_dim = time_dim
        if self.downsample_factor != int(self.downsample_factor):
            raise ValueError("Fractional downsampling not supported.")
        if self.downsample_factor < 1:
            raise ValueError(
                "Downsampling must be by a factor of 1 (no upsampling) or greater."
            )

    def __call__(self, video: torch.Tensor) -> torch.Tensor:
        video_length = video.size(self.time_dim)
        downsampled_indices = torch.arange(
            0, video_length, self.downsample_factor
        ).long()
        frames = torch.index_select(video, self.time_dim, downsampled_indices)
        return frames


class TemporalCrop(object):
    """Video transformation for performing "temporal cropping:" the sampling of
    a pre-defined number of clips, each with pre-defined length, from a full
    video. Can be used with Compose in torchvision.

    When used with TemporalDownsampling, it is highly recommended to put
    TemporalCrop after TemporalDownsampling. Since TemporalCrop can change the
    number of dimensions in the output tensor, due to clip selection, it is in
    fact recommended to put this transform at the end of a video transformation
    pipeline.

    Arguments:
        n_clips (int): the number of clips that should be sampled.
        clip_length (int): the length of each clip (in the number of frames)
        time_dim (int): the index of the time dimension of the video
        clip_spacing (Optional; default "equal"): how to choose starting locations
            for sampling clips. Keyword "equal" means that clip starting locations
            are sampled from each 1/n_clips segment of the video. The other option,
            "anywhere", places no restrictions on where clip starting locations
            can be sampled.
        padding_mode: (Optional; default "loop"): behavior if a requested clip
            length would result a clip exceeding the end of the video. Keyword
            "loop" results in a wrap-around to the start of the video. The other
            option, "freeze", repeats the final frame until the requested clip
            length is achieved.
        sample_starting_location: (Optional; default False): whether to sample a
            starting location (usually used for training) for a clip. Can be used
            in tandem with "equal" during training to sample clips with random
            starting locations distributed across time. Redundant if `clip_spacing`
            is "anywhere".
        stack_clips: (Optional; default True): whether to stack clips in a new
            dimension (used in 3D action recognition backbones), or stack clips by
            concatenating across the time dimension (used in 2D action recognition
            backbones). Output shape if True is (n_clips, *video_shape). If False,
            the output shape has the same number of dimensions as the original
            video, but the time dimension is extended by a factor of n_clips.


    Examples:

        # Create a VideoCell from "/path/to/video.mp4" with time in dimension one,
        sampling 10 clips each of length 16, sampling clips equally across the video
        >>> cell = VideoCell("/path/to/video.mp4",
            time_dim=1,
            transform=TemporalCrop(10, 16, time_dim=1)
            )
        # output shape: (10, n_channels, 16, H, W)

        # Create a VideoCell from "/path/to/video.mp4" with time in dimension one,
        sampling 8 clips each of length 8, sampling clips from arbitrary video
        locations and freezing the last frame if a clip exceeds the video length
        >>> cell = VideoCell("/path/to/video.mp4",
            time_dim=1,
            transform=TemporalCrop(8, 8, time_dim=1, clip_spacing="anywhere",
            padding_mode="freeze")
            )
        # output shape: (8, n_channels, 8, H, W)

        # Create a VideoCell from "/path/to/video.mp4" with time in dimension one,
        sampling one frame from each third of the video, concatenating the frames
        in the time dimension
        >>> cell = VideoCell("/path/to/video.mp4",
            time_dim=1,
            transform=TemporalCrop(1, 3, time_dim=1, clip_spacing="equal",
                sample_starting_location=True, stack_clips=False)
            )
        # output shape: (n_channels, 3, H, W)

    Note that time_dim in the TemporalDownsampling call must match the the time_dim
    in the VideoCell constructor!
    """

    def __init__(
        self,
        n_clips: int,
        clip_length: int,
        time_dim: Optional[int] = 1,
        clip_spacing: Optional[str] = "equal",
        padding_mode: Optional[str] = "loop",
        sample_starting_location: Optional[bool] = False,
        stack_clips: Optional[bool] = True,
    ):
        self.n_clips = n_clips
        self.clip_length = clip_length
        self.time_dim = time_dim
        self.clip_spacing = clip_spacing
        self.padding_mode = padding_mode
        self.stack_clips = stack_clips
        self.sample_starting_location = sample_starting_location
        if clip_length != int(clip_length):
            raise ValueError("Clip length (# of frames per clip) must be an integer")
        if clip_length <= 0:
            raise ValueError(
                "Clip length (# of frames per clip) must be a positive integer."
            )
        if n_clips != int(n_clips):
            raise ValueError("Number of clips is not an integer")
        if n_clips <= 0:
            raise ValueError("Number of clips must be a positive integer")
        assert clip_spacing in ["equal", "anywhere"]
        assert padding_mode in ["loop", "freeze"]

    def _get_sampling_boundaries(
        self, video_length: int, clip_number: int
    ) -> Tuple[int, int]:
        if self.clip_spacing == "equal":
            start = int(clip_number / self.n_clips * video_length)
            end = int((clip_number + 1) / self.n_clips * video_length)
        else:  # self.clip_spacing == "anywhere"
            start = 0
            end = video_length
        return start, end

    def _build_indices(self, start: int, length: int) -> torch.LongTensor:
        vanilla_indices = torch.arange(start, start + self.clip_length)
        if vanilla_indices.max().item() >= length:
            if self.padding_mode == "loop":
                vanilla_indices %= length
            else:  # self.padding_mode == "freeze":
                vanilla_indices[vanilla_indices >= length] = length - 1
        return vanilla_indices

    def __call__(self, video: torch.Tensor) -> torch.Tensor:
        video_length = video.size(self.time_dim)
        clips = []
        for clip_number in range(self.n_clips):
            start, end = self._get_sampling_boundaries(video_length, clip_number)
            if self.sample_starting_location:
                first_frame = random.randint(start, end)
            else:
                first_frame = start
            indices = self._build_indices(first_frame, video_length)
            clip = torch.index_select(video, self.time_dim, indices)
            clips.append(clip)
        if self.stack_clips:  # new dim for clips (n_clips, n_channels, duration, h, w)
            all_clips = torch.stack(clips, dim=0)
        else:  # concat clips in time dimension (n_channels, n_clips * duration, h, w)
            all_clips = torch.cat(clips, dim=self.time_dim)

        return all_clips



import os
from typing import Dict, Mapping

import ujson as json

import meerkat as mk

# TODO (Sabri): Add support for downloading the data.
# For the time being, relevant files should be downloaded to the directory here
# https://visualgenome.org/VGViz/explore


def build_visual_genome_dps(
    dataset_dir: str, write: bool = False
) -> Dict[str, mk.DataPanel]:
    dps = {}
    print("Loading objects and attributes...")
    dps.update(_build_object_dps(dataset_dir))

    print("Loading images...")
    dps.update(_build_image_dp(dataset_dir=dataset_dir))

    print("Loading relationships...")
    dps.update(_build_relationships_dp(dataset_dir=dataset_dir))

    if write:
        write_visual_genome_dps(dps, dataset_dir=dataset_dir)
    return dps


def read_visual_genome_dps(dataset_dir: str) -> Dict[str, mk.DataPanel]:
    return {
        key: mk.DataPanel.read(os.path.join(dataset_dir, f"{key}.mk"))
        for key in ["attributes", "relationships", "objects", "images"]
    }


def write_visual_genome_dps(dps: Mapping[str, mk.DataPanel], dataset_dir: str):
    for key, dp in dps.items():
        dp.write(os.path.join(dataset_dir, f"{key}.mk"))


def _build_object_dps(dataset_dir: str):
    with open(os.path.join(dataset_dir, "attributes.json")) as f:
        objects = json.load(f)

    objects_dp = []  # create one table for objects
    attributes_dp = []  # create one table of attributes
    for image in objects:
        for obj in image["attributes"]:
            obj["image_id"] = image["image_id"]

            # all names are length 1
            names = obj.pop("names")
            obj["name"] = names[0]

            # add attributes to the table
            attributes = obj.pop("attributes", None)
            if attributes is not None:
                for attribute in attributes:
                    attributes_dp.append(
                        {"object_id": obj["object_id"], "attribute": attribute}
                    )

            # the vast majority of objects (99.7%) have 0 or 1 synonym in their
            # synset, so we only consider the first synonym to keep things simple
            synset = obj.pop("synsets")
            obj["syn_name"] = synset[0] if len(synset) > 0 else ""

            objects_dp.append(obj)

    return {
        "objects": mk.DataPanel(objects_dp),
        "attributes": mk.DataPanel(attributes_dp),
    }


def _build_image_dp(dataset_dir: str):
    with open(os.path.join(dataset_dir, "image_data.json")) as f:
        images = json.load(f)
    image_dp = mk.DataPanel(images)
    image_dp.remove_column("coco_id")
    image_dp.remove_column("flickr_id")

    image_dp["local_path"] = dataset_dir + (image_dp["url"].str.split("rak248")).apply(
        lambda x: x[-1]
    )

    image_dp["image"] = mk.ImageColumn(image_dp["local_path"])

    return {"images": image_dp}


def _build_relationships_dp(dataset_dir: str):

    with open(os.path.join(dataset_dir, "relationships.json")) as f:
        relationships = json.load(f)

    rel_dp = []
    for image in relationships:
        image_id = image["image_id"]
        for r in image["relationships"]:
            object_synset = r["object"]["synsets"]
            subject_synset = r["subject"]["synsets"]

            rel_dp.append(
                {
                    "image_id": image_id,
                    "predicate": r["predicate"],
                    "subject_object_id": r["subject"]["object_id"],
                    "subject_name": r["subject"]["name"]
                    if "name" in r["subject"]
                    else r["subject"]["names"][0],
                    "subject_syn": subject_synset[0] if len(subject_synset) > 0 else "",
                    "object_object_id": r["object"]["object_id"],
                    "object_name": r["object"]["name"]
                    if "name" in r["object"]
                    else r["object"]["names"][0],
                    "object_syn": object_synset[0] if len(object_synset) > 0 else "",
                }
            )
    rel_dp = mk.DataPanel(rel_dp)
    return {"relationships": rel_dp}



from __future__ import annotations

import itertools
from typing import List, Sequence, Union

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm

from meerkat.columns.tensor_column import TensorColumn
from meerkat.datapanel import DataPanel
from meerkat.ml.prediction_column import (
    ClassificationOutputColumn,
    _ClassifierOutputType,
)
from meerkat.tools.lazy_loader import LazyLoader

cv2 = LazyLoader("cv2")

Columnable = Union[Sequence, np.ndarray, pd.Series, torch.Tensor]


# TODO(Priya): Make probs, preds method return SegmentationOutputColumn type
class SegmentationOutputColumn(ClassificationOutputColumn):
    def __init__(
        self,
        logits: Columnable = None,
        probs: Columnable = None,
        preds: Columnable = None,
        num_classes: int = None,
        *args,
        **kwargs,
    ):
        super(SegmentationOutputColumn, self).__init__(
            logits=logits,
            probs=probs,
            preds=preds,
            num_classes=num_classes,
            *args,
            **kwargs,
        )

    def binarymask(
        self, class_index: int
    ) -> TensorColumn:  # TODO(Priya): Check column type

        if self.num_classes > 2 and class_index is None:
            raise ValueError("Provide class_index in case of multi-class segmentation")

        if self.num_classes == 2:
            # Binary mask is same as predictions
            mask = TensorColumn(
                self.data
                if self._ctype == _ClassifierOutputType.PREDICTION
                else self.predictions().data
            )
        else:
            # Convert to predictions if required
            preds = (
                self.data
                if self._ctype == _ClassifierOutputType.PREDICTION
                else self.predictions().data
            )
            mask = TensorColumn(torch.where(preds == class_index, 1, 0))

        return mask

    @staticmethod
    def rle2mask(
        dataset: DataPanel,
        input_columns: List[str],  # TODO(Priya): Support multiple RLE columns?
        orig_dim,
        resize_dim=None,
        to_nan: bool = False,
        batch_size: int = 32,
    ) -> TensorColumn:

        masks = []

        for batch in tqdm(
            dataset[input_columns].batch(batch_size),
            total=(len(dataset) // batch_size + int(len(dataset) % batch_size != 0)),
        ):

            batch_masks = _convert_rle2mask(
                batch, input_columns, orig_dim, resize_dim, to_nan
            )

            masks = list(itertools.chain(masks, batch_masks))

        masks_col = TensorColumn(masks)
        dataset.add_column(f"Binary Mask (from {input_columns[0]})", masks_col)

        return masks_col


def _convert_rle2mask(
    batch: DataPanel,
    input_columns: List[str],
    orig_dim,
    resize_dim=None,
    to_nan: bool = False,
):

    """Convert run length encoding (RLE) to 2D binary mask.

    Args:
    batch (DataPanel): DataPanel.
    input_columns: List of columns containing Run Length Encodings
    orig_dim (Tuple[int]): Shape of the image.
    resize_dim (Tuple[int]): Shape to resize to.
      Resizing is done with cubic interporlation.
    to_nan (bool, optional): Convert 0s to np.nan.

    Returns:
    List[np.ndarray]: List of np.ndarray containing binary mask
    """

    # TODO(Priya): Support for multiple input_columns?
    rle_data = batch[input_columns[0]].data
    height, width = orig_dim
    masks = []

    for rle in rle_data:
        mask = np.zeros(width * height)
        if rle != "-1":
            array = np.asarray([int(x) for x in rle.split()])
            starts = array[0::2]
            lengths = array[1::2]
            current_position = 0

            for index, start in enumerate(starts):
                current_position += start
                mask[current_position : current_position + lengths[index]] = 1
                current_position += lengths[index]
            mask = mask.reshape(width, height)

        if resize_dim is not None:
            mask = cv2.resize(mask, resize_dim, interpolation=cv2.INTER_CUBIC)
        if to_nan:
            mask[mask == 0] = np.nan

        masks.append(mask)

    return masks



from __future__ import annotations

import itertools
from typing import Sequence

from tqdm import tqdm

from meerkat.columns.list_column import ListColumn
from meerkat.columns.tensor_column import TensorColumn
from meerkat.tools.lazy_loader import LazyLoader

ops = LazyLoader("torchvision.ops")


class InstancesColumn(ListColumn):
    def __init__(self, data: Sequence = None, *args, **kwargs):

        super(InstancesColumn, self).__init__(data=data, *args, **kwargs)

    def num_instances(self) -> TensorColumn:
        """Returns the number of instances for each image."""

        data_col = TensorColumn([len(instance) for instance in self])

        return data_col

    def get_field(self, field: str, batch_size: int = 32) -> ListColumn:
        """Returns a speific field of the Instances object.

        Returns:
            ListColumn: List of the specified field from Instances object.
        """

        data = []
        for batch in tqdm(
            self.batch(batch_size),
            total=(len(self) // batch_size + int(len(self) % batch_size != 0)),
        ):
            batch_data = [instance.get_fields()[field] for instance in batch]
            data = list(itertools.chain(data, batch_data))

        data_col = ListColumn(data)

        return data_col

    def nms(
        self,
        iou_threshold: float,
        batch_size: int = 32,
    ) -> InstancesColumn:
        """Returns the instances retained after NMS for each image.

        Returns:
            ListColumn: Contains tensors of shape (N,4) representing the boxes retained
                        after NMS where N is the number of boxes.
        """

        data = []
        for batch in tqdm(
            self.batch(batch_size),
            total=(len(self) // batch_size + int(len(self) % batch_size != 0)),
        ):

            batch_data = [
                instance.get_fields()["pred_boxes"].tensor[
                    ops.batched_nms(
                        boxes=instance.get_fields()["pred_boxes"].tensor,
                        scores=instance.get_fields()["scores"],
                        idxs=instance.get_fields()["pred_classes"],
                        iou_threshold=iou_threshold,
                    )
                ]
                for instance in batch
            ]
            data = list(itertools.chain(data, batch_data))

        data_col = ListColumn(data)
        return data_col



import os
from pathlib import Path
from typing import Sequence, Union

PathLikeType = Union[str, Path, os.PathLike]
PathLike = (str, Path, os.PathLike)


class FileMixin:
    """Mixin for adding in single filepath."""

    # TODO(karan): this also actually works on dirs, so rename

    def __init__(self, filepath: Union[str, Path], *args, **kwargs):
        super(FileMixin, self).__init__(*args, **kwargs)

        if isinstance(filepath, str):
            filepath = Path(filepath)

        # Assign the path
        self.filepath = filepath

    def __getattr__(self, item):
        try:
            return getattr(self.filepath, item)
        except AttributeError:
            raise AttributeError(f"Attribute {item} not found.")


class PathsMixin:
    """Mixin for adding in generic paths."""

    def __init__(
        self,
        paths: Union[PathLikeType, Sequence[PathLikeType]],
        *args,
        **kwargs,
    ):
        super(PathsMixin, self).__init__(*args, **kwargs)

        if isinstance(paths, PathLike):
            paths = [Path(paths)]
        elif not isinstance(paths, str) and isinstance(paths, Sequence):
            paths = [Path(p) for p in paths]
        else:
            raise NotImplementedError

        # Assign the path
        # TODO: make this a property
        self.paths = paths

    def __getattr__(self, item):
        try:
            return [getattr(p, item) for p in self.paths]
        except AttributeError:
            raise AttributeError(f"Attribute {item} not found.")



import os
import tarfile

import pandas as pd
from torchvision.datasets.utils import download_url

import meerkat as mk

VERSION_TO_URL = {
    "full": "https://s3.amazonaws.com/fast-ai-imageclas/imagenette2.tgz",
    "320px": "https://s3.amazonaws.com/fast-ai-imageclas/imagenette2-320.tgz",
    "160px": "https://s3.amazonaws.com/fast-ai-imageclas/imagenette2-160.tgz",
}

ID_TO_WORDS = {
    "n02979186": "cassette player",
    "n03417042": "garbage truck",
    "n01440764": "tench",
    "n02102040": "english springer spaniel",
    "n03028079": "church",
    "n03888257": "parachute",
    "n03394916": "french horn",
    "n03000684": "chainsaw",
    "n03445777": "golf ball",
    "n03425413": "gas pump",
}

ID_TO_IDX = {
    "n02979186": 482,
    "n03417042": 569,
    "n01440764": 0,
    "n02102040": 217,
    "n03028079": 497,
    "n03888257": 701,
    "n03394916": 566,
    "n03000684": 491,
    "n03445777": 574,
    "n03425413": 571,
}


def download_imagenette(
    download_dir, version="160px", overwrite: bool = False, return_df: bool = False
):
    """Download Imagenette dataset.

    Args:
        download_dir (str): The directory path to save to.
        version (str, optional): Imagenette version.
            Choices: ``"full"``, ``"320px"``, ``"160px"``.
        overwrite (bool, optional): If ``True``, redownload the dataset.
        return_df (bool, optional): If ``True``, return a ``pd.DataFrame``.

    Returns:
        Union[str, pd.DataFrame]: If ``return_df=True``, returns a pandas DataFrame.
            Otherwise, returns the directory path where the data is stored.

    References:
        https://github.com/fastai/imagenette
    """
    tar_path = os.path.join(download_dir, os.path.basename(VERSION_TO_URL[version]))
    dir_path = os.path.splitext(tar_path)[0]
    csv_path = os.path.join(dir_path, "imagenette.csv")
    if not overwrite and os.path.isfile(csv_path):
        return pd.read_csv(csv_path) if return_df else dir_path

    if overwrite or not os.path.exists(dir_path):
        download_url(
            url=VERSION_TO_URL[version],
            root=download_dir,
        )
        print("Extracting tar archive, this may take a few minutes...")
        tar = tarfile.open(tar_path)
        tar.extractall(download_dir)
        tar.close()
        os.remove(tar_path)
    else:
        print(f"Directory {dir_path} already exists. Skipping download.")

    # build dataframe
    df = pd.read_csv(os.path.join(dir_path, "noisy_imagenette.csv"))
    df["label_id"] = df["noisy_labels_0"]
    df["label"] = df["label_id"].replace(ID_TO_WORDS)
    df["label_idx"] = df["label_id"].replace(ID_TO_IDX)
    df["split"] = df["is_valid"].replace({False: "train", True: "valid"})
    df["img_path"] = df.path.apply(lambda x: os.path.join(dir_path, x))
    df[["img_path", "label", "label_id", "label_idx", "split"]].to_csv(
        csv_path, index=False
    )
    return df if return_df else dir_path


def build_imagenette_dp(
    dataset_dir: str,
    download: bool = False,
    version: str = "160px",
) -> mk.DataPanel:
    """Build DataPanel for the Imagenette dataset.

    Args:
        download_dir (str): The directory path to save to or load from.
        version (str, optional): Imagenette version.
            Choices: ``"full"``, ``"320px"``, ``"160px"``.
        overwrite (bool, optional): If ``True``, redownload the datasets.

    Returns:
        mk.DataPanel: A DataPanel corresponding to the dataset.

    References:
        https://github.com/fastai/imagenette
    """
    if download:
        df: pd.DataFrame = download_imagenette(
            dataset_dir, version=version, overwrite=False, return_df=True
        )
    else:
        csv_path = os.path.join(dataset_dir, "imagenette.csv")
        if not os.path.isfile(csv_path):
            raise ValueError("Imagenette is not downloaded. Pass `download=True`.")
        df = pd.read_csv(csv_path)

    dp = mk.DataPanel.from_pandas(df)
    dp["img"] = mk.ImageColumn.from_filepaths(dp["img_path"])
    return dp



import os
import subprocess
from typing import Dict

import pandas as pd

import meerkat as mk


def build_imagenet_dps(
    dataset_dir: str, download: bool = False
) -> Dict[str, mk.DataPanel]:

    if download:
        curr_dir = os.getcwd()
        os.makedirs(dataset_dir, exist_ok=True)
        os.chdir(dataset_dir)
        subprocess.run(
            args=[
                "kaggle competitions download "
                "-c imagenet-object-localization-challenge",
            ],
            shell=True,
            check=True,
        )
        subprocess.run(["unzip imagenet-object-localization-challenge.zip"])
        subprocess.run(
            ["tar", "-xzvf", "imagenet_object_localization_patched2019.tar.gz"]
        )
        os.chdir(curr_dir)

    # load training data
    paths = pd.read_csv(
        os.path.join(dataset_dir, "ILSVRC/ImageSets/CLS-LOC/train_cls.txt"),
        delimiter=" ",
        names=["path", "idx"],
    )["path"]
    train_df = paths.str.extract(r"(?P<synset>.*)/(?P<image_id>.*)")

    train_df["path"] = paths.apply(
        lambda x: os.path.join(dataset_dir, "ILSVRC/Data/CLS-LOC/train", f"{x}.JPEG")
    )
    train_df["split"] = "train"

    # load validation data
    valid_df = pd.read_csv(
        "/home/common/datasets/imagenet/LOC_val_solution.csv"
    ).rename(columns={"ImageId": "image_id"})
    valid_df["synset"] = valid_df["PredictionString"].str.split(" ", expand=True)[0]
    valid_df["path"] = valid_df["image_id"].apply(
        lambda x: os.path.join(dataset_dir, "ILSVRC/Data/CLS-LOC/val", f"{x}.JPEG")
    )
    valid_df["split"] = "valid"

    dp = mk.DataPanel.from_pandas(
        pd.concat([train_df, valid_df.drop(columns="PredictionString")])
    )
    dp["image"] = mk.ImageColumn.from_filepaths(dp["path"])

    # mapping from synset to english
    with open("/home/common/datasets/imagenet/LOC_synset_mapping.txt") as f:
        lines = f.read().splitlines()
    df = (
        pd.Series(lines)
        .str.split(" ", expand=True, n=1)
        .rename(columns={0: "synset", 1: "name"})
    )
    dp = dp.merge(mk.DataPanel.from_pandas(df), how="left", on="synset")

    return dp



from __future__ import annotations

import abc
import base64
import logging
from io import BytesIO
from typing import Sequence

import cytoolz as tz
from yaml.representer import Representer

import meerkat as mk
from meerkat.columns.abstract import AbstractColumn
from meerkat.mixins.cloneable import CloneableMixin
from meerkat.tools.lazy_loader import LazyLoader

PIL = LazyLoader("PIL")


Representer.add_representer(abc.ABCMeta, Representer.represent_name)


logger = logging.getLogger(__name__)


# Q. how to handle collate and materialize here? Always materialized but only sometimes
# may want to collate (because collate=True should return a batch-style object, while
# collate=False should return a Column style object).


class ListColumn(AbstractColumn):
    def __init__(
        self,
        data: Sequence = None,
        *args,
        **kwargs,
    ):
        if data is not None:
            data = list(data)
        super(ListColumn, self).__init__(data=data, *args, **kwargs)

    @classmethod
    def from_list(cls, data: Sequence):
        return cls(data=data)

    def batch(
        self,
        batch_size: int = 1,
        drop_last_batch: bool = False,
        collate: bool = True,
        *args,
        **kwargs,
    ):
        for i in range(0, len(self), batch_size):
            if drop_last_batch and i + batch_size > len(self):
                continue
            if collate:
                yield self.collate(self[i : i + batch_size])
            else:
                yield self[i : i + batch_size]

    def _repr_cell(self, index) -> object:
        return self[index]

    def _get_formatter(self) -> callable:
        if not mk.config.DisplayOptions.show_images:
            return None

        max_image_width = mk.config.DisplayOptions.max_image_width
        max_image_height = mk.config.DisplayOptions.max_image_height

        def _image_base64(im):
            with BytesIO() as buffer:
                im.save(buffer, "jpeg")
                return base64.b64encode(buffer.getvalue()).decode()

        def _image_formatter(cell):
            im = cell
            if isinstance(im, PIL.Image.Image):
                im.thumbnail((max_image_width, max_image_height))
                return f'<img src="data:image/jpeg;base64,{_image_base64(im)}">'
            else:
                return repr(cell)

        return _image_formatter

    @classmethod
    def concat(cls, columns: Sequence[ListColumn]):
        data = list(tz.concat([c.data for c in columns]))
        if issubclass(cls, CloneableMixin):
            return columns[0]._clone(data=data)
        return cls.from_list(data)

    def is_equal(self, other: AbstractColumn) -> bool:
        return (self.__class__ == other.__class__) and self.data == other.data



import os
import warnings
from typing import List

from numpy.lib.format import open_memmap

from meerkat.columns.tensor_column import TensorColumn
from meerkat.datapanel import DataPanel
from meerkat.ml.activation import ActivationOp
from meerkat.tools.lazy_loader import LazyLoader

pl = LazyLoader("pytorch_lightning")


class ActivationCallback(pl.callbacks.Callback):
    def __init__(
        self,
        target_module: str,
        val_len: int,
        logdir: None,
        mmap: bool = False,
    ):
        """Callback to store model activations during validation step.

        Args:
            target_module (str): the name of the submodule of `model` (i.e. an
                intermediate layer) that outputs the activations we'd like to extract.
                For nested submodules, specify a path separated by "." (e.g.
                `ActivationCachedOp(model, "block4.conv")`).
            val_len (int): Number of inputs in the validation dataset
            logdir (str): Directory to store the activation datapanels.
                Not required if trainer.log_dir is to be used
            mmap (bool): If true, activations are stored as memmapped numpy arrays
        """

        super().__init__()
        self.target_module = target_module
        self.val_len = val_len
        self.logdir = logdir
        self.mmap = mmap
        self.shape = None  # Shape of activations

        if self.mmap:
            warnings.warn(
                "Activations will be stored as numpy array when using memmapping."
            )

    def on_init_start(self, trainer):
        if self.logdir is None:
            self.logdir = trainer.log_dir

    def on_validation_epoch_start(self, trainer, pl_module):
        if not trainer.running_sanity_check:
            self.activation_op = ActivationOp(
                pl_module, self.target_module, pl_module.device
            )
            self.writer = TensorColumn.get_writer(mmap=self.mmap)

    def on_validation_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
    ):
        # TODO(Priya): Check if skipping sanity check is fine
        if not trainer.running_sanity_check:
            activations = self.activation_op.extractor.activation.cpu().detach()

            # Use the first batch for setup
            if batch_idx == 0:
                if self.mmap:
                    shape = (self.val_len, *activations.shape[1:])
                    self.shape = shape

                    # TODO(Priya): File name format
                    file = f"activations_{self.target_module}_{trainer.current_epoch}"
                    self.writer.open(os.path.join(self.logdir, file), shape=shape)

                else:
                    self.writer.open()
            self.writer.write(activations)

    def on_validation_epoch_end(self, trainer, pl_module):
        if not trainer.running_sanity_check:
            activations = {f"activation_{self.target_module}": self.writer.finalize()}
            activations = DataPanel.from_batch(activations)

            if not self.mmap:
                file = f"activations_{self.target_module}_{trainer.current_epoch}.mk"
                activations.write(os.path.join(self.logdir, file))


def load_activations(
    target_module: str,
    logdir: str,
    epochs: List[int],
    mmap: bool = False,
    mmap_mode: str = "r",
) -> DataPanel:

    activations_dp = None

    for epoch in epochs:

        path = os.path.join(logdir, f"activations_{target_module}_{epoch}")
        if not mmap:
            path += ".mk"

        if not os.path.exists(path):
            raise ValueError(f"{path} does not exist.")

        if mmap:
            activations = open_memmap(path, mode=mmap_mode)
        else:
            activations = DataPanel.read(path)[f"activation_{target_module}"]

        if activations_dp is None:
            activations_dp = DataPanel(
                {f"activation_{target_module}_{epoch}": activations}
            )
        else:
            activations_dp.add_column(
                f"activation_{target_module}_{epoch}", activations
            )

    return activations_dp



from __future__ import annotations

import base64
import logging
import os
import warnings
from io import BytesIO
from typing import Collection, Mapping, Sequence, Union

import numpy as np
import yaml

import meerkat as mk
from meerkat.cells.abstract import AbstractCell
from meerkat.columns.abstract import AbstractColumn
from meerkat.datapanel import DataPanel
from meerkat.errors import ConcatWarning
from meerkat.tools.lazy_loader import LazyLoader

PIL = LazyLoader("PIL")


logger = logging.getLogger(__name__)


class LambdaCell(AbstractCell):
    def __init__(
        self,
        fn: callable = None,
        data: any = None,
    ):
        self.fn = fn
        self._data = data

    @property
    def data(self) -> object:
        """Get the data associated with this cell."""
        return self._data

    def get(self, *args, **kwargs):
        if isinstance(self.data, AbstractCell):
            return self.fn(self.data.get())
        elif isinstance(self.data, Mapping):
            return self.fn(
                {
                    k: v.get() if isinstance(v, AbstractCell) else v
                    for k, v in self.data.items()
                }
            )
        else:
            return self.fn(self.data)

    def __eq__(self, other):
        return (
            (other.__class__ == self.__class__)
            and (self.data == other.data)
            and (self.fn == other.fn)
        )

    def __repr__(self):
        name = getattr(self.fn, "__qualname__", repr(self.fn))
        return f"LambdaCell(fn={name})"


class LambdaColumn(AbstractColumn):
    def __init__(
        self,
        data: Union[DataPanel, AbstractColumn],
        fn: callable = None,
        output_type: type = None,
        *args,
        **kwargs,
    ):
        super(LambdaColumn, self).__init__(data.view(), *args, **kwargs)
        if fn is not None:
            self.fn = fn
        self._output_type = output_type

    def _set(self, index, value):
        raise ValueError("Cannot setitem on a `LambdaColumn`.")

    def fn(self, data: object):
        """Subclasses like `ImageColumn` should be able to implement their own
        version."""
        raise NotImplementedError

    def _create_cell(self, data: object) -> LambdaCell:
        return LambdaCell(fn=self.fn, data=data)

    def _get_cell(self, index: int, materialize: bool = True):
        if materialize:
            return self.fn(self._data._get(index, materialize=True))
        else:
            return self._create_cell(data=self._data._get(index, materialize=False))

    def _get_batch(self, indices: np.ndarray, materialize: bool = True):
        if materialize:
            # if materializing, return a batch (by default, a list of objects returned
            # by `.get`, otherwise the batch format specified by `self.collate`)
            data = self.collate(
                [self._get_cell(int(i), materialize=True) for i in indices]
            )
            if self._output_type is not None:
                data = self._output_type(data)
            return data
        else:
            return self._data.lz[indices]

    def _get(self, index, materialize: bool = True, _data: np.ndarray = None):
        index = self._translate_index(index)
        if isinstance(index, int):
            if _data is None:
                _data = self._get_cell(index, materialize=materialize)
            return _data

        elif isinstance(index, np.ndarray):
            # support for blocks
            if _data is None:
                _data = self._get_batch(index, materialize=materialize)
            if materialize:
                # materialize could change the data in unknown ways, cannot clone
                return self.__class__.from_data(data=_data)
            else:
                return self._clone(data=_data)

    @classmethod
    def _state_keys(cls) -> Collection:
        return super()._state_keys() | {"fn", "_output_type"}

    @staticmethod
    def concat(columns: Sequence[LambdaColumn]):
        for c in columns:
            if c.fn != columns[0].fn:
                warnings.warn(
                    ConcatWarning("Concatenating LambdaColumns with different `fn`.")
                )
                break

        return columns[0]._clone(mk.concat([c._data for c in columns]))

    def _write_data(self, path):
        # TODO (Sabri): avoid redundant writes in dataframes
        return self.data.write(os.path.join(path, "data"))

    def is_equal(self, other: AbstractColumn) -> bool:
        if other.__class__ != self.__class__:
            return False
        if self.fn != other.fn:
            return False

        return self.data.is_equal(other.data)

    @staticmethod
    def _read_data(path: str):
        meta = yaml.load(
            open(os.path.join(path, "data", "meta.yaml")),
            Loader=yaml.FullLoader,
        )
        if issubclass(meta["dtype"], AbstractColumn):
            return AbstractColumn.read(os.path.join(path, "data"))
        else:
            return DataPanel.read(os.path.join(path, "data"))

    def _repr_cell(self, idx):
        return self.lz[idx]

    def _get_formatter(self) -> callable:
        if not mk.config.DisplayOptions.show_images:
            return None

        max_image_width = mk.config.DisplayOptions.max_image_width
        max_image_height = mk.config.DisplayOptions.max_image_height

        def _image_base64(im):
            with BytesIO() as buffer:
                im.save(buffer, "jpeg")
                return base64.b64encode(buffer.getvalue()).decode()

        def _image_formatter(cell):
            im = cell.get()
            if isinstance(im, PIL.Image.Image):
                im.thumbnail((max_image_width, max_image_height))
                return f'<img src="data:image/jpeg;base64,{_image_base64(im)}">'
            else:
                return repr(cell)

        return _image_formatter



import os
from pathlib import Path


class DisplayOptions:
    max_rows: int = 10
    show_images: bool = True

    max_image_height: int = 128
    max_image_width: int = 128


class ContribOptions:
    download_dir: str = os.path.join(Path.home(), ".meerkat/datasets")



"""Import common classes."""
# flake8: noqa

from meerkat.logging.utils import initialize_logging

initialize_logging()

from meerkat.cells.abstract import AbstractCell
from meerkat.cells.spacy import LazySpacyCell, SpacyCell
from meerkat.cells.volume import MedicalVolumeCell
from meerkat.columns.abstract import AbstractColumn
from meerkat.columns.arrow_column import ArrowArrayColumn
from meerkat.columns.cell_column import CellColumn
from meerkat.columns.image_column import ImageColumn
from meerkat.columns.lambda_column import LambdaCell, LambdaColumn
from meerkat.columns.list_column import ListColumn
from meerkat.columns.numpy_column import NumpyArrayColumn
from meerkat.columns.pandas_column import PandasSeriesColumn
from meerkat.columns.spacy_column import SpacyColumn
from meerkat.columns.tensor_column import TensorColumn
from meerkat.columns.video_column import VideoColumn
from meerkat.columns.volume_column import MedicalVolumeColumn
from meerkat.contrib.registry import datasets
from meerkat.datapanel import DataPanel
from meerkat.ops.concat import concat
from meerkat.ops.merge import merge
from meerkat.provenance import provenance



import os
from typing import Callable, Collection, Optional

import torch

from meerkat import AbstractCell
from meerkat.contrib.video_corruptions.utils import stderr_suppress
from meerkat.tools.lazy_loader import LazyLoader

cv2 = LazyLoader("cv2")
F = LazyLoader("torchvision.transforms.functional")


class VideoCell(AbstractCell):
    """Interface for loading video data.

    Examples:

    # Load a video from "/path/to/video.mp4", where the time dimension has index one
    >>> cell = VideoCell("/path/to/video.mp4", time_dim=1)
    """

    def __init__(
        self,
        filepath: str,
        transform: Optional[Callable] = None,
        time_dim: Optional[int] = 1,
    ):
        super().__init__()
        self.filepath = filepath
        self.time_dim = time_dim
        self.transform = transform

    def _read_all_frames(self):
        if not os.path.isfile(self.filepath):
            raise ValueError(f"{self.filepath} is not a valid file!")
        cap = cv2.VideoCapture(self.filepath)
        frames = []
        ret = True
        while ret:
            ret, img = cap.read()
            try:
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            except cv2.error:
                break
            img = F.to_tensor(img)
            frames.append(img)
        cap.release()
        frames = torch.stack(frames, dim=self.time_dim)
        return frames

    def get(self):
        with stderr_suppress():
            frames = self._read_all_frames()  # TODO: support different decoders
        if self.transform is not None:
            frames = self.transform(frames)
        return frames

    @classmethod
    def _state_keys(cls) -> Collection:
        return {"filepath", "time_dim", "transform"}



from __future__ import annotations

import logging
from typing import Collection, Sequence

from meerkat.columns.lambda_column import LambdaColumn
from meerkat.columns.pandas_column import PandasSeriesColumn

logger = logging.getLogger(__name__)


class ReportColumn(LambdaColumn):
    def __init__(
        self,
        data: Sequence[str] = None,
        transform: callable = None,
        loader: callable = None,
        *args,
        **kwargs,
    ):
        super(ReportColumn, self).__init__(
            PandasSeriesColumn.from_data(data), *args, **kwargs
        )
        self.loader = self.default_loader if loader is None else loader
        self.transform = transform

    def fn(self, filepath: str):
        image = self.loader(filepath)
        if self.transform is not None:
            image = self.transform(image)
        return image

    @classmethod
    def from_filepaths(
        cls,
        filepaths: Sequence[str],
        loader: callable = None,
        transform: callable = None,
        *args,
        **kwargs,
    ):
        return cls(
            data=filepaths,
            loader=loader,
            transform=transform,
            *args,
            **kwargs,
        )

    @classmethod
    def default_loader(cls, filepath):
        with open(filepath) as f:
            return f.read()

    @classmethod
    def _state_keys(cls) -> Collection:
        return (super()._state_keys() | {"transform", "loader"}) - {"fn"}



from __future__ import annotations

import abc
import functools
import logging
import numbers
import os
import shutil
from mmap import mmap
from typing import Callable, Sequence

import numpy as np
import pandas as pd
import torch
from yaml.representer import Representer

from meerkat.block.abstract import BlockView
from meerkat.block.numpy_block import NumpyBlock
from meerkat.columns.abstract import AbstractColumn
from meerkat.writers.concat_writer import ConcatWriter

Representer.add_representer(abc.ABCMeta, Representer.represent_name)

logger = logging.getLogger(__name__)


def getattr_decorator(fn: Callable):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        out = fn(*args, **kwargs)
        if isinstance(out, np.ndarray):
            return NumpyArrayColumn(out)
        else:
            return out

    return wrapper


class NumpyArrayColumn(
    AbstractColumn,
    np.lib.mixins.NDArrayOperatorsMixin,
):

    block_class: type = NumpyBlock

    def __init__(
        self,
        data: Sequence,
        *args,
        **kwargs,
    ):
        if isinstance(data, BlockView):
            if not isinstance(data.block, NumpyBlock):
                raise ValueError(
                    "Cannot create `NumpyArrayColumn` from a `BlockView` not "
                    "referencing a `NumpyBlock`."
                )
        elif not isinstance(data, np.memmap):
            data = np.asarray(data)
        super(NumpyArrayColumn, self).__init__(data=data, *args, **kwargs)

    # TODO (sabri): need to support str here
    _HANDLED_TYPES = (np.ndarray, numbers.Number)

    def __array_ufunc__(self, ufunc: np.ufunc, method, *inputs, **kwargs):
        out = kwargs.get("out", ())
        for x in inputs + out:
            # Only support operations with instances of _HANDLED_TYPES.
            # Use ArrayLike instead of type(self) for isinstance to
            # allow subclasses that don't override __array_ufunc__ to
            # handle ArrayLike objects.
            if not isinstance(x, self._HANDLED_TYPES + (NumpyArrayColumn,)) and not (
                # support for at index
                method == "at"
                and isinstance(x, list)
            ):
                return NotImplemented

        # Defer to the implementation of the ufunc on unwrapped values.
        inputs = tuple(x.data if isinstance(x, NumpyArrayColumn) else x for x in inputs)

        if out:
            kwargs["out"] = tuple(
                x.data if isinstance(x, NumpyArrayColumn) else x for x in out
            )
        result = getattr(ufunc, method)(*inputs, **kwargs)

        if type(result) is tuple:
            # multiple return values
            return tuple(type(self)(x) for x in result)
        elif method == "at":
            # no return value
            return None
        else:
            # one return value
            return self._clone(data=result)

    def __getattr__(self, name):
        try:
            out = getattr(object.__getattribute__(self, "data"), name)
            if isinstance(out, Callable):
                return getattr_decorator(out)
            else:
                return out
        except AttributeError:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            )

    @classmethod
    def from_array(cls, data: np.ndarray, *args, **kwargs):
        return cls(data=data, *args, **kwargs)

    def _set_batch(self, indices, values):
        self._data[indices] = values

    def _get(self, index, materialize: bool = True):
        index = NumpyBlock._convert_index(index)
        data = self._data[index]
        if self._is_batch_index(index):
            # only create a numpy array column
            return self._clone(data=data)
        else:
            return data

    def _copy_data(self) -> object:
        return self._data.copy()

    def _view_data(self) -> object:
        return self._data

    @property
    def is_mmap(self):
        # important to check if .base is a python mmap object, since a view of a mmap
        # is also a memmap object, but should not be symlinked or copied
        return isinstance(self.data, np.memmap) and isinstance(self.data.base, mmap)

    def _write_data(self, path: str, link: bool = True) -> None:
        path = os.path.join(path, "data.npy")
        # important to check if .base is a python mmap object, since a view of a mmap
        # is also a memmap object, but should not be symlinked
        if self.is_mmap:
            if link:
                os.symlink(self.data.filename, path)
            else:
                shutil.copy(self.data.filename, path)
        else:
            np.save(path, self.data)

    @staticmethod
    def _read_data(path: str, mmap=False, *args, **kwargs) -> np.ndarray:
        data_path = os.path.join(path, "data.npy")

        if mmap:
            return np.load(data_path, mmap_mode="r")
        return np.load(data_path)

    @classmethod
    def concat(cls, columns: Sequence[NumpyArrayColumn]):
        data = np.concatenate([c.data for c in columns])
        return columns[0]._clone(data=data)

    def is_equal(self, other: AbstractColumn) -> bool:
        if other.__class__ != self.__class__:
            return False
        return np.array_equal(self.data, other.data, equal_nan=True)

    @classmethod
    def get_writer(cls, mmap: bool = False, template: AbstractColumn = None):
        if mmap:
            from meerkat.writers.numpy_writer import NumpyMemmapWriter

            return NumpyMemmapWriter()
        else:
            return ConcatWriter(template=template, output_type=NumpyArrayColumn)

    def _repr_cell(self, index) -> object:
        if len(self.shape) > 1:
            return f"np.ndarray(shape={self.shape[1:]})"
        else:
            return self[index]

    def to_tensor(self) -> torch.Tensor:
        """Use `column.to_tensor()` instead of `torch.tensor(column)`, which is
        very slow."""
        # TODO (Sabri): understand why `torch.tensor(column)` is so slow
        return torch.tensor(self.data)

    def to_pandas(self) -> pd.Series:
        if len(self.shape) == 1:
            return pd.Series(self.data)
        else:
            # can only create a 1-D series
            return super().to_pandas()

    @classmethod
    def from_npy(
        cls,
        path,
        mmap_mode=None,
        allow_pickle=False,
        fix_imports=True,
        encoding="ASCII",
    ):
        data = np.load(
            path,
            mmap_mode=mmap_mode,
            allow_pickle=allow_pickle,
            fix_imports=fix_imports,
            encoding=encoding,
        )
        return cls(data)



from __future__ import annotations

import os
import shutil
from dataclasses import dataclass
from mmap import mmap
from typing import Hashable, Sequence, Tuple, Union

import numpy as np
import torch

from meerkat.block.ref import BlockRef
from meerkat.errors import ConsolidationError

from .abstract import AbstractBlock, BlockIndex, BlockView


class NumpyBlock(AbstractBlock):
    @dataclass(eq=True, frozen=True)
    class Signature:
        dtype: np.dtype
        nrows: int
        shape: Tuple[int]
        klass: type
        mmap: Union[bool, int]

    def __init__(self, data, *args, **kwargs):
        super(NumpyBlock, self).__init__(*args, **kwargs)
        if len(data.shape) <= 1:
            raise ValueError(
                "Cannot create a `NumpyBlock` from data with less than 2 axes."
            )
        self.data = data

    @property
    def signature(self) -> Hashable:
        return self.Signature(
            klass=NumpyBlock,
            # don't want to consolidate any mmaped blocks
            mmap=id(self) if isinstance(self.data, np.memmap) else False,
            nrows=self.data.shape[0],
            shape=self.data.shape[2:],
            dtype=self.data.dtype,
        )

    def _get_data(self, index: BlockIndex, materialize: bool = True) -> np.ndarray:
        return self.data[:, index]

    @classmethod
    def from_column_data(cls, data: np.ndarray) -> Tuple[NumpyBlock, BlockView]:
        """[summary]

        Args:
            data (np.ndarray): [description]
            names (Sequence[str]): [description]

        Raises:
            ValueError: [description]

        Returns:
            Tuple[NumpyBlock, Mapping[str, BlockIndex]]: [description]
        """
        if len(data.shape) == 1:
            data = np.expand_dims(data, axis=1)
            block_index = 0
        elif data.shape[1] == 1:
            block_index = slice(0, 1)
        else:
            block_index = slice(0, data.shape[1])

        block = cls(data)
        return BlockView(block=block, block_index=block_index)

    @classmethod
    def _consolidate(
        cls,
        block_refs: Sequence[BlockRef],
    ) -> BlockRef:
        offset = 0
        new_indices = {}
        columns = {}
        to_concat = []
        for block_ref in block_refs:
            for name, col in block_ref.items():
                # keep track of all the columns in the block_refs
                if name in columns:
                    raise ConsolidationError(
                        "Cannot consolidate two block refs containing the same column."
                    )
                columns[name] = col

                # add block and compute new indices
                block_index = col._block_index
                if isinstance(block_index, slice):
                    block_view = col._block.data[:, block_index]
                    new_indices[name] = slice(
                        # need to update slice offset and remove step
                        offset,
                        block_view.shape[1] + offset,
                        1,
                    )
                elif isinstance(block_index, int):
                    # keep block axis
                    block_view = col._block.data[:, block_index : block_index + 1]
                    new_indices[name] = offset
                to_concat.append(block_view)
                offset += block_view.shape[1]

        block = cls(np.concatenate(to_concat, axis=1))

        # create columns
        new_columns = {
            name: columns[name]._clone(data=block[block_index])
            for name, block_index in new_indices.items()
        }
        return BlockRef(block=block, columns=new_columns)

    @staticmethod
    def _convert_index(index):
        if torch.is_tensor(index):
            # need to convert to numpy for boolean indexing
            return index.numpy()
        return index

    def _get(
        self, index, block_ref: BlockRef, materialize: bool = True
    ) -> Union[BlockRef, dict]:
        index = self._convert_index(index)
        # TODO: check if they're trying to index more than just the row dimension
        data = self.data[index]
        if isinstance(index, int):
            # if indexing a single row, we do not return a block manager, just a dict
            return {
                name: data[col._block_index] for name, col in block_ref.columns.items()
            }
        block = self.__class__(data)
        columns = {
            name: col._clone(data=block[col._block_index])
            for name, col in block_ref.columns.items()
        }
        # note that the new block may share memory with the old block
        return BlockRef(block=block, columns=columns)

    @property
    def is_mmap(self):
        # important to check if .base is a python mmap object, since a view of a mmap
        # is also a memmap object, but should not be symlinked or copied
        return isinstance(self.data, np.memmap) and isinstance(self.data.base, mmap)

    def _write_data(self, path: str, link: bool = True):
        path = os.path.join(path, "data.npy")
        if self.is_mmap:
            if link:
                os.symlink(self.data.filename, path)
            else:
                shutil.copy(self.data.filename, path)
        else:
            np.save(path, self.data)

    @staticmethod
    def _read_data(path: str, mmap: bool = False):
        data_path = os.path.join(path, "data.npy")

        if mmap:
            return np.load(data_path, mmap_mode="r")
        return np.load(data_path)



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping, Sequence, Union

if TYPE_CHECKING:
    from meerkat.block.abstract import AbstractBlock
    from meerkat.columns.abstract import AbstractColumn


class BlockRef(Mapping):
    def __init__(self, columns: Mapping[str, AbstractColumn], block: AbstractBlock):
        self.columns: Mapping[str, AbstractColumn] = columns
        self.block: AbstractBlock = block

    def __getitem__(self, index: Union[str, Sequence[str]]):
        if isinstance(index, str):
            return self.columns[index]
        else:
            return self.__class__(
                columns={col: self.columns[col].view() for col in index},
                block=self.block,
            )

    def __delitem__(self, key):
        self.columns.pop(key)

    def __len__(self):
        return len(self.columns)

    def __contains__(self, value):
        return value in self.columns

    def __iter__(self):
        return iter(self.columns)

    @property
    def block_indices(self):
        return {name: col._block_index for name, col in self.columns.items()}

    def apply(self, method_name: str = "_get", *args, **kwargs):
        # apply method to the block
        return getattr(self.block, method_name)(*args, **kwargs, block_ref=self)

    def update(self, block_ref: BlockRef):
        if id(block_ref.block) != id(self.block):
            raise ValueError(
                "Can only update BlockRef with another BlockRef pointing "
                "to the same block."
            )
        self.columns.update(block_ref.columns)



"""EntityDataPanel Class."""
from __future__ import annotations

import logging
from typing import Any, Dict, List, Sequence, Tuple, Union

import datasets
import numpy as np
import torch

from meerkat import DataPanel, ListColumn, NumpyArrayColumn, TensorColumn
from meerkat.ml import EmbeddingColumn

logger = logging.getLogger(__name__)


class EntityDataPanel(DataPanel):
    def __init__(
        self,
        data: Union[dict, list, datasets.Dataset] = None,
        columns: List[str] = None,
        embedding_columns: List[str] = None,
        index_column: str = None,
        **kwargs,
    ):
        """
        EntityDataPanel: A data panel with "atomic" unit semantics.
        This means each row represents a unique "atom" or thing.
        Each row has an index column that is used in merges.
        Further, each row can get one or more embedding columns
        'associated with it. This allows for Embedding
        operations such as nearest neighbor search.

        Args:
            columns: all column names
            embedding_columns: embedding columns in all columns
            index_column: index column
        """
        super().__init__(
            data=data,
            **kwargs,
        )
        if len(self.columns) > 0:
            self._embedding_columns = embedding_columns if embedding_columns else []

            self._check_columns_unique(self._embedding_columns)
            self._check_columns_exist(self._embedding_columns)
            for c in self._embedding_columns:
                self._cast_to_embedding(c)

            self._index_column = index_column if index_column else None
            if not self._index_column:
                self._add_ent_index()
            else:
                self._check_index_unique()
            self._check_columns_exist([self._index_column])
            self._index_to_rowid = {idx: i for i, idx in enumerate(self.index)}
        else:  # Initializing empty EntityDataPanel DP - needed when loading
            self._embedding_columns = []
            self._index_column = None
            self._index_to_rowid = {}

    def _check_columns_exist(self, columns: List[str]):
        """Check that every column in `columns` exists."""
        for col in columns:
            assert col in self.columns, f"{col} is not a valid column."

    def _check_columns_unique(self, columns: List[str]):
        """Checks that all columns are unique."""
        assert len(columns) == len(set(columns))

    @classmethod
    def from_datapanel(
        cls,
        datapanel: DataPanel,
        embedding_columns: List[str] = None,
        index_column: str = None,
    ):
        """Returns EntityDataPanel DP from standard DP."""
        return cls(
            data=datapanel.data,
            embedding_columns=embedding_columns,
            index_column=index_column,
        )

    def to_datapanel(self, klass: type = None):
        """Casts to a DataPanel.

        Args:
            klass (type): If specified, casts to this class.
                This class must be a subclass of :ref:`meerkat.datapanel.DataPanel`.
                If not specified, defaults to :ref:`meerkat.datapanel.DataPanel`.

        Returns:
            DataPanel: The datapanel.
        """
        if klass is None:
            klass = DataPanel
        elif not issubclass(klass, DataPanel):
            raise ValueError("`klass` must be a subclass of DataPanel")
        return klass.from_batch({k: self[k] for k in self.columns})

    @property
    def index(self):
        """Returns index column."""
        return self[self.index_column]

    @property
    def embedding_columns(self):
        """Returns _visible_ embedding columns."""
        return [e for e in self._embedding_columns if e in self.columns]

    @property
    def index_column(self):
        """Returns index column."""
        return self._index_column

    def _check_index_unique(self):
        assert len(self.index) == len(
            set(self.index)
        ), "Index must be unique and hashable"

    def icontain(self, idx: Any):
        """Checks if idx in the index column or not."""
        return idx in self._index_to_rowid

    def iget(self, idx: Any):
        """Gets the row given the entity index."""
        idx_col_type = type(next(iter(self._index_to_rowid.keys())))
        if not isinstance(idx, idx_col_type):
            raise ValueError(
                f"Query ({type(idx)}) must be the same type as the "
                f"index column ({idx_col_type}) of the data"
            )
        assert idx in self._index_to_rowid, f"{idx} not in index set"
        row_idx = self._index_to_rowid[idx]
        if self.index.visible_rows is not None:
            # Map from original index into visible row index (inverse of remap index)
            try:
                row_idx = int(np.where(self.index.visible_rows == row_idx)[0][0])
            except IndexError:
                raise IndexError(f"{idx} not in data")
        return self[row_idx]

    def _add_ent_index(self):
        """Add an integer index to the dataset.

        Not using index from DataPanel as ids need to be integers to
        serve as embedding row indices
        """
        # TODO: Why do you have a string index...feels weird and expensive.
        # with row indexes due to filtering.
        self.add_column("_ent_index", [i for i in range(len(self))])
        self._index_column = "_ent_index"

    def _cast_to_embedding(self, name):
        """Cast column to Embedding column if not already."""
        if not isinstance(self._data[name], EmbeddingColumn):
            self._data[name] = EmbeddingColumn(self._data[name])

    def add_embedding_column(
        self,
        name: str,
        embedding: Union[np.ndarray, torch.tensor, EmbeddingColumn],
        index_to_rowid: Dict[Any, int] = None,
        overwrite: bool = False,
    ):
        """Adds embedding column to data.

        If index_to_rowid provided, maps DP index column to rowid of
        embedding.
        """
        if index_to_rowid is not None:
            permutation = [index_to_rowid[i] for i in self.index]
            embedding = embedding[permutation]
        assert len(embedding) == len(
            self
        ), "Length of embedding needs to be the same as data"
        self.add_column(name, embedding, overwrite)
        self._cast_to_embedding(name)
        # If name is duplicate, overwrite must be true
        # otherwise ``add_column`` would fail
        if name not in self._embedding_columns:
            self._embedding_columns.append(name)

    def remove_column(self, column: str) -> None:
        """Remove column.

        Assert index column is not removed
        """
        assert column != self.index_column, "Can't remove an index column"
        super(EntityDataPanel, self).remove_column(column)
        if column in self._embedding_columns:
            self._embedding_columns.remove(column)

    def append(
        self,
        dp: EntityDataPanel,
        axis: Union[str, int] = "rows",
        suffixes: Tuple[str] = None,
        overwrite: bool = False,
    ) -> EntityDataPanel:
        """Append an EntityDataPanel."""
        if axis == 0 or axis == "rows":
            # append new rows
            return super(EntityDataPanel, self).append(dp, axis, suffixes, overwrite)
        elif axis == 1 or axis == "columns":
            # append new columns; must make sure the entities are in the same order
            # data property takes `visible_rows` into account
            if not (self.index.data == dp.index.data).all():
                raise ValueError(
                    "Can only append along axis 1 (columns) if the entity indexes match"
                )
            new_embedding_cols = self._get_merged_embedding_columns(
                dp, overwrite, suffixes
            )
            # Save the new index column for saving EntityDataPanel
            new_index_column = self.index_column
            if self.index_column in dp.columns and not overwrite:
                new_index_column += suffixes[0]
            ret = super(EntityDataPanel, self).append(dp, axis, suffixes, overwrite)
            ret._embedding_columns = new_embedding_cols
            ret._index_column = new_index_column
            return ret
        else:
            raise ValueError("DataPanel `axis` must be either 0 or 1.")

    def merge(
        self,
        right: "EntityDataPanel",
        how: str = "inner",
        left_on: str = None,
        sort: bool = False,
        suffixes: Sequence[str] = ("_x", "_y"),
        validate=None,
        keep_indexes: bool = False,
    ):
        """Perform merge of two EntityDPs.

        By default we merge both on their index columns. We do allow for
        changing the join column of the left side
        """
        left_on = left_on if left_on is not None else self.index_column
        new_embedding_cols = self._get_merged_embedding_columns(
            right, overwrite=False, suffixes=suffixes
        )
        # If the left index column is in the right set of columns,
        # and the left index column won't stay the same because it's the
        # the joining column, then the index column will change to
        # have a suffix
        new_index_column = self.index_column
        if self.index_column in right.columns:
            # Column will stay the same if it's the joining
            # column of both left and right
            if not (
                self.index_column == left_on and self.index_column == right.index_column
            ):
                assert suffixes is not None and len(suffixes) == 2, (
                    "The index column is the same as rights columns. "
                    "Must provide suffixes"
                )
                new_index_column += suffixes[0]
        # Merge calls append and EntityDataPanel append ensures
        # indexes are aligned. When merging, we adopt the index of
        # the left, making it okay that indexes are unaligned. To avoid
        # this, we cast to DataPanel.
        ret = self.to_datapanel().merge(
            right.to_datapanel(),
            how=how,
            left_on=left_on,
            right_on=right.index_column,
            sort=sort,
            suffixes=suffixes,
            validate=validate,
            keep_indexes=False,
        )
        return EntityDataPanel.from_datapanel(
            datapanel=ret,
            embedding_columns=new_embedding_cols,
            index_column=new_index_column,
        )

    def _get_merged_embedding_columns(self, dp, overwrite, suffixes):
        """In order to create EntityDataPanel post merge/append, we need to
        have the embedding columns.

        This calculates the new EntityDataPanel embedding columns in the
        case of suffix changes.
        """
        new_embedding_cols = self.embedding_columns + dp.embedding_columns
        # Capture the new embedding columns before merging
        if len(set(self.embedding_columns).intersection(dp.embedding_columns)) > 0:
            if overwrite:
                new_embedding_cols = self.embedding_columns + [
                    c for c in dp.embedding_columns if c not in self.embedding_columns
                ]
            else:
                assert (
                    suffixes is not None and len(suffixes) == 2
                ), "Suffixes must be tuple of len 2 when columns share names"
                new_embedding_cols = (
                    [c for c in self.embedding_columns if c not in dp.embedding_columns]
                    + [
                        c
                        for c in dp.embedding_columns
                        if c not in self.embedding_columns
                    ]
                    + [
                        c + suffixes[0]
                        for c in self.embedding_columns
                        if c in dp.embedding_columns
                    ]
                    + [
                        c + suffixes[1]
                        for c in dp.embedding_columns
                        if c in self.embedding_columns
                    ]
                )
        return new_embedding_cols

    def convert_entities_to_ids(
        self, column: Union[ListColumn, TensorColumn, NumpyArrayColumn]
    ):
        """Maps column of entity idx to their row ids for the embeddings.

        Used in data prep before training.
        """

        def recursive_map(seq):
            if isinstance(seq, (np.ndarray, torch.Tensor, list)):
                return [recursive_map(item) for item in seq]
            else:
                # TODO: handle UNK entity ids
                return self._index_to_rowid[seq]

        assert isinstance(column, (ListColumn, TensorColumn, NumpyArrayColumn)), (
            "We only support DataPanel list column types "
            "(ListColumn, TensorColumn, NumpyArrayColumn)"
        )
        return column.map(lambda x: recursive_map(x))

    def most_similar(
        self,
        query: Any,
        k: int,
        query_embedding_column=None,
        search_embedding_column=None,
    ):
        """Returns most similar entities distinct from query."""
        assert k < len(self), "k must be less than the total number of entities"
        if query_embedding_column is None:
            query_embedding_column = self.embedding_columns[0]
        self._check_columns_exist([query_embedding_column])
        if search_embedding_column is None:
            search_embedding_column = query_embedding_column
        else:
            self._check_columns_exist([search_embedding_column])
            assert (
                self[query_embedding_column].shape[-1]
                == self[search_embedding_column].shape[-1]
            ), "Length of search embedding needs to match query embedding"
        # Will noop if already exists
        self[search_embedding_column].build_faiss_index(overwrite=False)

        emb_query = self.iget(query)[query_embedding_column].numpy().reshape(1, -1)
        dist, sims = self[search_embedding_column].search(emb_query, k + 1)
        # May or may not return the emb_query. If the embeddings are not unique,
        # we must selectively remove the query in the answer
        sims = sims[0][sims[0] != self._index_to_rowid[query]]
        sims = sims[:k]
        return self[sims]

    @classmethod
    def _state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return {
            "_embedding_columns",
            "_index_column",
        }

    def _clone_kwargs(self) -> Dict[str, Any]:
        default_kwargs = super()._clone_kwargs()
        default_kwargs.update({"index_column": self.index_column})
        return default_kwargs



import os
import subprocess

import pandas as pd

import meerkat as mk

ATTRIBUTES = [
    "5_o_clock_shadow",
    "arched_eyebrows",
    "attractive",
    "bags_under_eyes",
    "bald",
    "bangs",
    "big_lips",
    "big_nose",
    "black_hair",
    "blond_hair",
    "blurry",
    "brown_hair",
    "bushy_eyebrows",
    "chubby",
    "double_chin",
    "eyeglasses",
    "goatee",
    "gray_hair",
    "heavy_makeup",
    "high_cheekbones",
    "male",
    "mouth_slightly_open",
    "mustache",
    "narrow_eyes",
    "no_beard",
    "oval_face",
    "pale_skin",
    "pointy_nose",
    "receding_hairline",
    "rosy_cheeks",
    "sideburns",
    "smiling",
    "straight_hair",
    "wavy_hair",
    "wearing_earrings",
    "wearing_hat",
    "wearing_lipstick",
    "wearing_necklace",
    "wearing_necktie",
    "young",
]


def get_celeba(dataset_dir: str, download: bool = False):
    """Build the dataframe by joining on the attribute, split and identity
    CelebA CSVs."""
    if download:
        download_celeba(dataset_dir=dataset_dir)
    df = build_celeba_df(dataset_dir=dataset_dir)
    dp = mk.DataPanel.from_pandas(df)
    dp["image"] = mk.ImageColumn.from_filepaths(
        filepaths=dp["img_path"], base_dir=dataset_dir
    )
    return dp


def download_celeba(dataset_dir: str):
    # if not os.path.exists(dataset_dir):
    #     CelebA(os.path.split(dataset_dir)[:-1][0], download=True)
    if os.path.exists(dataset_dir):
        return
    curr_dir = os.getcwd()
    os.makedirs(dataset_dir, exist_ok=True)
    os.chdir(dataset_dir)
    subprocess.run(
        args=["kaggle datasets download " "-d jessicali9530/celeba-dataset"],
        shell=True,
        check=True,
    )
    subprocess.run(["unzip", "-q", "celeba-dataset.zip"])
    os.chdir(curr_dir)


def build_celeba_df(dataset_dir: str):
    """Build the dataframe by joining on the attribute, split and identity
    CelebA CSVs."""
    identity_df = pd.read_csv(
        os.path.join(dataset_dir, "identity_CelebA.txt"),
        delim_whitespace=True,
        header=None,
        names=["file", "identity"],
    )
    attr_df = pd.read_csv(
        os.path.join(dataset_dir, "list_attr_celeba.csv"),
        index_col=0,
    )
    attr_df.columns = pd.Series(attr_df.columns).apply(lambda x: x.lower())
    attr_df = ((attr_df + 1) // 2).rename_axis("file").reset_index()

    celeb_df = identity_df.merge(attr_df, on="file", validate="one_to_one")

    celeb_df["img_path"] = celeb_df.file.apply(
        lambda x: os.path.join("img_align_celeba/img_align_celeba", x)
    )

    split_df = pd.read_csv(os.path.join(dataset_dir, "list_eval_partition.csv"))
    split_df["split"] = split_df["partition"].replace(
        {0: "train", 1: "valid", 2: "test"}
    )
    celeb_df = celeb_df.merge(
        split_df[["image_id", "split"]], left_on="file", right_on="image_id"
    )
    return celeb_df



import logging
import os
import re
import subprocess
from typing import Iterable

import google.auth
import pandas as pd
from google.cloud import bigquery, bigquery_storage, storage
from PIL import Image
from pydicom.filereader import dcmread

import meerkat as mk

from .gcs import GCSImageColumn
from .modules import TABLES
from .reports import ReportColumn

logger = logging.getLogger(__name__)


def build_mimic_dp(
    dataset_dir: str,
    gcp_project: str,
    tables: Iterable[str] = None,
    excluded_tables: Iterable[str] = None,
    reports: bool = False,
    cxr_dicom: bool = True,
    cxr_jpg: bool = True,
    split: bool = True,
    download_jpg: bool = False,
    download_resize: int = None,
    write: bool = False,
) -> mk.DataPanel:
    """Builds a `DataPanel` for accessing data from the MIMIC-CXR database
    https://physionet.org/content/mimic-cxr/2.0.0/ The MIMIC-CXR database
    integrates chest X-ray imaging data with structured EHR data from Beth
    Israel Deaconess Medical Center. The full database has an uncompressed size
    of over 5 TB. This function quickly builds a `DataPanel` that can be used
    to explore, slice and download the database. Building the DataPanel takes.

    ~1 minute (when not downloading the radiology reports). The large CXR DICOM
    and JPEG files are not downloaded, but lazily pulled from Google Cloud
    Storage (GCS) only when they are accessed. This makes it possible to
    inspect and explore that data without downloading the full 5 TB.

    Note: model training will likely bottleneck on the GCS downloads, so it is
    recommended that you cache the JPEG images locally bfore training. This can be
    accomplished by setting a `writer` and running a map over the data.
    ```
        dp["jpg_img].writer = lambda path, img: x.save(path, img)
        dp["jpg_img].map(lambda x: True)
    ```
    The images will be saved in `dataset_dir`. This will take several hours for the full
    dataset. You can also slice down to a subset of the dataset before running the map.

    Each row corresponds to a single chest X-ray image (stored in both DICOM format and
    JPEG in the MIMIC database). Each row is uniquely identified by the "dicom_id"
    column. Note that a single chest X-ray study (identified by "study_id" column) may
    consist of multiple images and a single patient (identified by "subject_id" column)
    may have multiple studies in the database. The columns in the DataPanel can be
    grouped into four categories:
        1. (`PandasSeriesColumn`) Metadata and labels pulled from tables in the MIMIC-IV
            EHR database (e.g."pneumonia", "ethnicity", "view_position", "gender"). For
            more information on the tables see: https://mimic.mit.edu/docs/iv/modules/.
            For more information on the CheXpert labels see:
            https://physionet.org/content/mimic-cxr-jpg/2.0.0/
        2. (`GCSImageColumn`) DICOM and JPEG image files of the chest xrays.These
            columns do not actually hold the images, but instead lazily load from the
            GCS when they're indexed.
        3. (`ReportColumn`) The radiology reports for each exam are downloaded to disk,
            and lazily loaded when accessed.
    The arguments below can be used to specify which of the columns to include in the
    DataPanel.

    Args:
        dataset_dir (str): A local directory in which downloaded data will be cached.
        gcp_project (str): The Google Cloud Platform project that will be billed for
            downloads from the database (MIMIC has requester-pays enabled). If you do
            not have GCP project, see instructions for creating one here:
            https://cloud.google.com/resource-manager/docs/creating-managing-projects
        tables (Iterable[str], optional): A subset of ["patient", "admit", "labels",
            "dicom_meta"] specifying which tables to include in the DataPanel.
            Defaults to None, in which case all of the tables listed in
            "meerkat.contrib.mimic.TABLES" will be included.
        excluded_tables (Iterable[str], optional): A subset of ["patient", "admit",
            "labels", "dicom_meta"] specifying which tables to exclude from the
            DataPanel. Defaults to None, in which case none are excluded.
        reports (bool, optional): Download reports if they aren't already downloaded
            in `dataset_dir` and add a "report" column to the DataPanel. Defaults to
            False.
        cxr_dicom (bool, optional): Add a `GCSImagecolumn` called "cxr_dicom" to the
            DataPanel for the DICOM files for each image. Defaults to True.
        cxr_jpg (bool, optional):  Add a `GCSImagecolumn` called "cxr_jpg" to the
            DataPanel for the JPEG files for each image. Defaults to True.
        split (bool, optional): Add a "split" column with "train", "validate" and "test"
            splits. Defaults to True.
        download_jpg (bool, optional): Download jpegs for all the scans in the dataset
            to `dataset_dir`. Expect this to take several hours. Defaults to False.
        download_resize (bool, optional): Resize the images before saving them to disk.
            Defaults to None, in which case the images are not resized.
        write (bool, optiional): Write the datapanel to the directory.

    Returns:
        DataPanel: The MIMIC `DataPanel` with columns
    """
    os.environ["GOOGLE_CLOUD_PROJECT"] = gcp_project

    tables = set(TABLES.keys() if tables is None else tables)
    if excluded_tables is not None:
        tables -= set(excluded_tables)
    # must include the cxr_records table
    tables |= set(["cxr_records"])

    fields = [
        (
            f"{table}.{field[0]} AS {field[1]}"
            if isinstance(field, tuple)
            else f"{table}.{field}"
        )
        for table in tables
        for field in TABLES[table]["fields"]
        if table != "admit"
    ]
    query_str = f"""
            SELECT {','.join(fields)}
            FROM `physionet-data.mimic_cxr.record_list` cxr_records
            LEFT JOIN `physionet-data.mimic_cxr.study_list` cxr_studies
            ON cxr_records.study_id = cxr_studies.study_id
    """

    if "labels" in tables:
        query_str += """
            LEFT JOIN `physionet-data.mimic_cxr.chexpert` labels
                ON cxr_records.study_id = labels.study_id
        """

    if "patients" in tables:
        query_str += """
            LEFT JOIN `physionet-data.mimic_core.patients` patients
                ON cxr_records.subject_id = patients.subject_id
        """

    if "dicom_meta" in tables:
        query_str += """
            LEFT JOIN `physionet-data.mimic_cxr.dicom_metadata_string` AS dicom_meta
            ON cxr_records.dicom_id = dicom_meta.dicom
        """
    elif "admit" in tables:
        # need the StudyDate to include admission data
        query_str += """
            LEFT JOIN (SELECT StudyDate, dicom
                FROM `physionet-data.mimic_cxr.dicom_metadata_string`) AS meta
            ON cxr_records.dicom_id = meta.dicom
        """

    print(f"Querying MIMIC database: `gcp_project`={gcp_project}, `tables`={tables}.")
    df = query_mimic_db(query_str, gcp_project=gcp_project)

    if "admit" in tables:
        # joining in admissions data is more complicated because the study_list table
        # does not include an index into the admissions table
        # instead we match each study to the first admission for which the study date
        # falls between the admission and discharge dates
        fields = ["admit." + field for field in TABLES["admit"]["fields"]] + [
            "subject_id"
        ]

        admit_df = query_mimic_db(
            query_str=f"""
                SELECT {','.join(fields)}
                FROM `physionet-data.mimic_core.admissions` admit
            """,
            gcp_project=gcp_project,
        )
        admit_df = df[["subject_id", "StudyDate", "study_id"]].merge(
            admit_df, on="subject_id"
        )
        study_date = pd.to_datetime(admit_df["StudyDate"])
        admit_df = admit_df[
            (study_date >= admit_df["admittime"].dt.date)
            & (study_date <= admit_df["dischtime"].dt.date)
        ]
        df = df.merge(
            admit_df.drop(columns=["StudyDate", "subject_id"]).drop_duplicates(
                subset="study_id"
            ),
            how="left",
            on="study_id",
        )

    # convert dicom metadata from str to float
    if "dicom_meta" in tables:
        for field in TABLES["dicom_meta"]["fields"]:
            try:
                df[field] = df[field].astype(float)
            except ValueError:
                # if we can't convert, just keep the field as str
                continue

    # convert to snake case
    df = df.rename(columns=lambda x: re.sub(r"(?<!^)(?<!_)(?=[A-Z])", "_", x).lower())

    print("Preparing DataPanel...")
    dp = mk.DataPanel.from_pandas(df)

    # add GCSImageColumn for the jpg version of the xrays
    if cxr_jpg:
        paths = pd.Series(dp["dicom_path"].data)
        dp["jpg_path"] = paths.str.split(".").str[0] + ".jpg"
        dp["cxr_jpg"] = GCSImageColumn.from_blob_names(
            blob_names=dp["jpg_path"],
            bucket_name="mimic-cxr-jpg-2.0.0.physionet.org",
            project=gcp_project,
            loader=Image.open,
            local_dir=dataset_dir,
        )

    # add GCSImageColumn for the dicoms
    if cxr_dicom:
        dp["cxr_dicom"] = GCSImageColumn.from_blob_names(
            blob_names=dp["dicom_path"],
            bucket_name="mimic-cxr-2.0.0.physionet.org",
            project=gcp_project,
            loader=dcmread,
            local_dir=dataset_dir,
        )

    if reports:
        reports_dir = os.path.join(dataset_dir, "mimic-cxr-reports")
        if not os.path.exists(reports_dir):
            # download and unzip reports
            print("Downloading reports...")
            storage_client = storage.Client(project=gcp_project)
            bucket = storage_client.bucket(
                "mimic-cxr-2.0.0.physionet.org", user_project=gcp_project
            )
            filepath = os.path.join(dataset_dir, "mimic-cxr-reports.zip")
            bucket.blob("mimic-cxr-reports.zip").download_to_filename(filepath)
            subprocess.run(
                [
                    "unzip",
                    filepath,
                    "-d",
                    os.path.join(dataset_dir, "mimic-cxr-reports"),
                ]
            )
        dp["report"] = ReportColumn.from_filepaths(
            reports_dir + "/" + dp["report_path"]
        )

    if split:
        print("Downloading splits...")
        storage_client = storage.Client(project=gcp_project)
        bucket = storage_client.bucket(
            "mimic-cxr-jpg-2.0.0.physionet.org", user_project=gcp_project
        )
        filepath = os.path.join(dataset_dir, "mimic-cxr-2.0.0-split.csv.gz")
        bucket.blob("mimic-cxr-2.0.0-split.csv.gz").download_to_filename(
            filepath,
        )
        subprocess.run(["gunzip", filepath])

        dp = dp.merge(
            mk.DataPanel.from_csv(
                os.path.join(dataset_dir, "mimic-cxr-2.0.0-split.csv")
            )[["split", "dicom_id"]],
            how="left",
            on="dicom_id",
        )

    if download_jpg:
        dp = download_mimic_dp(dp, resize=download_resize)

    if write:
        dp.write(os.path.join(dataset_dir, "mimic.mk"))

    return dp


def query_mimic_db(query_str: str, gcp_project: str) -> pd.DataFrame:
    credentials, _ = google.auth.default(
        scopes=["https://www.googleapis.com/auth/cloud-platform"]
    )

    # Make clients.
    bqclient = bigquery.Client(
        credentials=credentials,
        project=gcp_project,
    )
    bqstorageclient = bigquery_storage.BigQueryReadClient(credentials=credentials)

    df = (
        bqclient.query(query_str)
        .result()
        .to_dataframe(bqstorage_client=bqstorageclient, progress_bar_type="tqdm")
    )

    return df


def download_mimic_dp(mimic_dp: mk.DataPanel, resize: int = None, **kwargs):
    col = mimic_dp["cxr_jpg"].view()
    dataset_dir = col.local_dir
    paths = mimic_dp["jpg_path"]
    if resize:
        paths = paths.apply(
            lambda x: os.path.join(
                dataset_dir, os.path.splitext(x)[0] + f"_{resize}" + ".jpg"
            )
        )

    def _write_resized(path, img):
        if resize is not None:
            img.thumbnail((resize, resize))
            root, ext = os.path.splitext(path)
            path = root + f"_{resize}" + ext
        img.save(path)

    col._skip_cache = True
    col.writer = _write_resized
    col.map(
        lambda x: True,
        num_workers=6,
        pbar=True,
    )

    mimic_dp[f"cxr_jpg_{resize}"] = mk.ImageColumn.from_filepaths(paths)
    return mimic_dp



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Hashable, List, Sequence, Union

import numpy as np
import pandas as pd
import pyarrow as pa
import torch

from meerkat.block.ref import BlockRef
from meerkat.columns.numpy_column import NumpyArrayColumn
from meerkat.columns.tensor_column import TensorColumn

from .abstract import AbstractBlock, BlockIndex, BlockView


class ArrowBlock(AbstractBlock):
    @dataclass(eq=True, frozen=True)
    class Signature:
        nrows: int
        klass: type
        # mmap: bool

    def __init__(self, data: pa.Table, *args, **kwargs):
        super(ArrowBlock, self).__init__(*args, **kwargs)
        self.data = data

    @property
    def signature(self) -> Hashable:
        return self.Signature(klass=ArrowBlock, nrows=len(self.data))

    def _get_data(self, index: BlockIndex) -> pa.Array:
        return self.data[index]

    @classmethod
    def from_column_data(cls, data: pa.Array) -> BlockView:
        data = pa.Table.from_pydict({"col": data})
        block = cls(data)
        return BlockView(block=block, block_index="col")

    @classmethod
    def from_block_data(cls, data: pa.Table) -> List[BlockView]:
        block = cls(data)
        return [
            BlockView(block=block, block_index=column) for column in data.column_names
        ]

    @classmethod
    def _consolidate(
        cls,
        block_refs: Sequence[BlockRef],
    ) -> BlockRef:
        table = pa.Table.from_pydict(
            # need to ignore index when concatenating
            {
                name: ref.block.data[col._block_index]
                for ref in block_refs
                for name, col in ref.items()
            }
        )
        block = cls(table)

        # pull out the block columns from all the block_refs
        columns = {}
        for ref in block_refs:
            columns.update(ref)

        new_columns = {
            name: col._clone(data=block[name]) for name, col in columns.items()
        }
        return BlockRef(block=block, columns=new_columns)

    @staticmethod
    def _convert_index(index):
        if isinstance(index, list):
            return np.array(index)
        if torch.is_tensor(index):
            # need to convert to numpy for boolean indexing
            return index.numpy()
        if isinstance(index, NumpyArrayColumn):
            return index.data
        if isinstance(index, TensorColumn):
            # need to convert to numpy for boolean indexing
            return index.data.numpy()
        if isinstance(index, pd.Series):
            # need to convert to numpy for boolean indexing
            return index.values
        from meerkat.columns.pandas_column import PandasSeriesColumn

        if isinstance(index, PandasSeriesColumn):
            return index.data.values

        return index

    def _get(
        self, index, block_ref: BlockRef, materialize: bool = True
    ) -> Union[BlockRef, dict]:
        index = self._convert_index(index)
        # TODO: check if they're trying to index more than just the row dimension

        if isinstance(index, slice) or isinstance(index, int):
            data = self.data[index]
        elif index.dtype == bool:
            data = self.data.filter(pa.array(index))
        else:
            data = self.data.take(index)

        if isinstance(index, int):
            # if indexing a single row, we do not return a block manager, just a dict
            return {
                name: data[col._block_index] for name, col in block_ref.columns.items()
            }
        block = self.__class__(data)

        columns = {
            name: col._clone(data=block[col._block_index])
            for name, col in block_ref.columns.items()
        }
        # note that the new block may share memory with the old block
        return BlockRef(block=block, columns=columns)

    @staticmethod
    def _write_table(path: str, table: pa.Table):
        # noqa E501, source: huggingface implementation https://github.com/huggingface/datasets/blob/92304b42cf0cc6edafc97832c07de767b81306a6/src/datasets/table.py#L50
        with open(path, "wb") as sink:
            writer = pa.RecordBatchStreamWriter(sink=sink, schema=table.schema)
            batches: List[pa.RecordBatch] = table.to_batches()
            for batch in batches:
                writer.write_batch(batch)
            writer.close()
            return sum(batch.nbytes for batch in batches)

    @staticmethod
    def _read_table(path: str, mmap: bool = False):
        if mmap:
            return pa.ipc.open_stream(pa.memory_map(path)).read_all()
        else:
            return pa.ipc.open_stream(pa.input_stream(path)).read_all()

    def _write_data(self, path: str):
        self._write_table(os.path.join(path, "data.arrow"), self.data)

    @staticmethod
    def _read_data(path: str, mmap: bool = False):
        return ArrowBlock._read_table(os.path.join(path, "data.arrow"), mmap=mmap)



import logging
import os
import pickle
from functools import partial

from tqdm import tqdm

import meerkat as mk

from .data_utils import (
    compute_file_tuples,
    compute_slice_matrix,
    compute_stanford_file_tuples,
    get_sz_labels,
    stanford_eeg_loader,
)

logger = logging.getLogger(__name__)


def build_eeg_dp(
    dataset_dir: str,
    raw_dataset_dir: str,
    splits=["train", "dev"],
    clip_len: int = 60,
    step_size: int = 1,
    stride: int = 60,
):
    """Builds a `DataPanel` for accessing EEG data.

    Currently only supports TUH dataset for seq-seq prediction.
    Future TODO: integrating stanford dataset with weak seq-seq labels

    Args:
        dataset_dir (str): A local directory where the preprocessed
            (h5) EEG data are stored
        raw_dataset_dir (str): A local directory where the original
            (edf) EEG data are stored
        clip_len (int): Number of seconds in an EEG clip
        step_size (int): Number of seconds in a single 'step'
        stride (int):  Number of seconds in the stride when extracting
            clips from signals
    """

    # retrieve paths of all edf files in the raw_dataset_dir
    edf_files = []
    for path, subdirs, files in os.walk(raw_dataset_dir):
        for name in files:
            if ".edf" in name:
                edf_files.append(os.path.join(path, name))

    data = []
    for split in splits:
        file_tuples = compute_file_tuples(
            raw_dataset_dir, dataset_dir, split, clip_len, stride
        )

        for (edf_fn, clip_idx, _) in tqdm(file_tuples, total=len(file_tuples)):
            filepath = [file for file in edf_files if edf_fn in file]
            filepath = filepath[0]
            file_id = edf_fn.split(".edf")[0]

            sequence_sz, binary_sz = get_sz_labels(
                edf_fn=filepath,
                clip_idx=int(clip_idx),
                time_step_size=step_size,
                clip_len=clip_len,
                stride=stride,
            )

            row_df = {
                "filepath": filepath,
                "file_id": file_id,
                "sequence_sz": sequence_sz,
                "binary_sz": binary_sz,
                "clip_idx": int(clip_idx),
                "h5_fn": os.path.join(dataset_dir, edf_fn.split(".edf")[0] + ".h5"),
                "split": split,
            }

            data.append(row_df)

    dp = mk.DataPanel(data)

    eeg_loader = partial(
        compute_slice_matrix, time_step_size=step_size, clip_len=clip_len, stride=stride
    )

    eeg_input_col = dp[["clip_idx", "h5_fn"]].to_lambda(fn=eeg_loader)

    dp.add_column(
        "eeg_input",
        eeg_input_col,
        overwrite=True,
    )

    return dp


def download_tusz(download_dir, version="1.5.2"):
    """Downloads the EEG Seizure TUH dataset (TUSZ)

    REQUIRED:
        1. Need to first registed at
            https://www.isip.piconepress.com/projects/tuh_eeg/html/downloads.shtml
        2. run download_tusz from python script or simply run the provided rsync
             command below in your terminal
        3. enter the provided password sent to your email after step (1)

    Args:
        download_dir (str): The directory path to save to.
        version (str, optional): Which version to download
    """

    src_pth = f"nedc@www.isip.piconepress.com:data/tuh_eeg_seizure/v{version}/"
    rsync_command = f"rsync -auxvL {src_pth} {download_dir}"
    print("Executing rsync command")
    os.system(rsync_command)


def build_stanford_eeg_dp(
    stanford_dataset_dir: str,
    lpch_dataset_dir: str,
    file_marker_dir: str,
    splits=["train", "dev"],
    reports_pth=None,
    clip_len: int = 60,
):
    """Builds a `DataPanel` for accessing EEG data.

    This is for accessing private stanford data.
    The stanford data is limited to specific researchers on IRB.
    No public directions on how to download them yet.
    Contact ksaab@stanford.edu for more information.

    Args:
        stanford_dataset_dir (str): A local dir where stanford EEG are stored
        lpch_dataset_dir (str): A local dir where the lpch EEG are stored
        file_marker_dir (str): A local dir where file markers are stored
        splits (list[str]): List of splits to load
        reports_pth (str): if not None, will load reports
        clip_len (int): Number of seconds in an EEG clip
    """

    # retrieve file tuples which is a list of
    # (eeg filepath, location of sz or -1 if no sz, split)
    file_tuples = compute_stanford_file_tuples(
        stanford_dataset_dir, lpch_dataset_dir, file_marker_dir, splits
    )
    data = []

    for (filepath, sz_loc, split) in file_tuples:
        row_df = {
            "filepath": filepath,
            "file_id": filepath.split("/")[-1].split(".eeghdf")[0],
            "binary_sz": sz_loc != -1,
            "sz_start_index": sz_loc,
            "split": split,
        }
        data.append(row_df)

    dp = mk.DataPanel(data)

    eeg_input_col = dp[["sz_start_index", "filepath", "split"]].to_lambda(
        fn=partial(stanford_eeg_loader, clip_len=clip_len)
    )

    dp.add_column(
        "eeg_input",
        eeg_input_col,
        overwrite=True,
    )

    if reports_pth:
        raw_reports_pth = os.path.join(reports_pth, "reports_unique_for_hl_mm.txt")
        raw_reports_dp = mk.DataPanel.from_csv(raw_reports_pth, sep="\t")

        parsed_reports_pth = os.path.join(reports_pth, "parsed_eeg_notes.dill")
        with open(parsed_reports_pth, "rb") as dill_f:
            parsed_reports = pickle.load(dill_f)

        doc_data = []
        for doc in parsed_reports:
            uuid = doc.doc_id
            mask_id = raw_reports_dp["note_uuid"] == uuid
            if mask_id.sum() == 1 and "findings" in doc.sections:
                file_id = raw_reports_dp[mask_id]["edf_file_name"][0].split(".edf")[0]
                findings = doc.sections["findings"]["text"]
                row_df = {"file_id": file_id, "findings": findings}
                doc_data.append(row_df)
        reports_dp = mk.DataPanel(doc_data)

        dp = dp.merge(reports_dp, how="left", on="file_id")

    return dp



from __future__ import annotations

import abc
import functools
import logging
import os
from typing import Callable, List, Mapping, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import torch
from yaml.representer import Representer

from meerkat.block.abstract import BlockView
from meerkat.block.tensor_block import TensorBlock
from meerkat.columns.abstract import AbstractColumn
from meerkat.mixins.cloneable import CloneableMixin
from meerkat.writers.concat_writer import ConcatWriter
from meerkat.writers.numpy_writer import NumpyMemmapWriter

Representer.add_representer(abc.ABCMeta, Representer.represent_name)

Columnable = Union[Sequence, np.ndarray, pd.Series, torch.Tensor]

logger = logging.getLogger(__name__)


def getattr_decorator(fn: Callable):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        out = fn(*args, **kwargs)
        if isinstance(out, torch.Tensor):
            if out.ndim == 0:
                return out.clone().detach()
            return TensorColumn(out)
        else:
            return out

    return wrapper


class TensorColumn(
    np.lib.mixins.NDArrayOperatorsMixin,
    AbstractColumn,
):
    block_class: type = TensorBlock

    def __init__(
        self,
        data: Sequence = None,
        *args,
        **kwargs,
    ):
        if isinstance(data, BlockView):
            if not isinstance(data.block, TensorBlock):
                raise ValueError(
                    "Cannot create `TensorColumn` from a `BlockView` not "
                    "referencing a `TensorBlock`."
                )
        elif data is not None and not isinstance(data, TensorColumn):
            if (
                isinstance(data, Sequence)
                and len(data) > 0
                and torch.is_tensor(data[0])
            ):
                # np.asarray supports a list of numpy arrays (it simply stacks them
                # before putting them into an array) but torch.as_tensor does not.
                # we want to support this for consistency and because it is important
                # for map
                data = torch.stack(data)
            data = torch.as_tensor(data)
        super(TensorColumn, self).__init__(data=data, *args, **kwargs)

    def __torch_function__(self, func, types, args=(), kwargs=None):
        def _process_arg(arg):
            if isinstance(arg, type(self)):
                return arg.data
            elif isinstance(arg, (List, Tuple)):
                # Specifically use list and tuple because these are
                # expected types for arguments in torch operations.
                return type(arg)([_process_arg(_a) for _a in arg])
            elif isinstance(arg, Mapping):
                # All mappings can be converted to dictionaries
                # when processed by torch operations.
                return {_k: _process_arg(_a) for _k, _a in arg.items()}
            else:
                return arg

        def _process_ret(ret):
            # This function may need to be refactored into an instance method
            # because the from_data implementation is different for each
            # class.
            if isinstance(ret, torch.Tensor):
                if ret.ndim == 0:
                    return ret.clone().detach()
                return self.from_data(ret)
            elif isinstance(ret, (List, Tuple)):
                return type(ret)([_process_arg(_a) for _a in ret])
            elif isinstance(ret, Mapping):
                return {_k: _process_arg(_a) for _k, _a in ret.items()}
            else:
                return ret

        if kwargs is None:
            kwargs = {}
        args = [_process_arg(a) for a in args]
        ret = func(*args, **kwargs)
        return _process_ret(ret)

    def __getattr__(self, name):
        try:
            out = getattr(object.__getattribute__(self, "data"), name)
            if isinstance(out, Callable):
                return getattr_decorator(out)
            else:
                return out
        except AttributeError:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            )

    def _get(self, index, materialize: bool = True):
        index = self.block_class._convert_index(index)

        data = self._data[index]
        if self._is_batch_index(index):
            # only create a numpy array column
            return self._clone(data=data)
        else:
            return data

    def _set_batch(self, indices, values):
        self._data[indices] = values

    @classmethod
    def concat(cls, columns: Sequence[TensorColumn]):
        data = torch.cat([c.data for c in columns])
        if issubclass(cls, CloneableMixin):
            return columns[0]._clone(data=data)
        return cls(data)

    @classmethod
    def get_writer(cls, mmap: bool = False, template: AbstractColumn = None):
        if mmap:
            return NumpyMemmapWriter()
        else:
            return ConcatWriter(template=template, output_type=TensorColumn)

    def _repr_cell(self, index) -> object:
        if len(self.shape) > 1:
            return f"torch.Tensor(shape={self.shape[1:]})"
        else:
            return self[index]

    @classmethod
    def from_data(cls, data: Union[Columnable, AbstractColumn]):
        """Convert data to an EmbeddingColumn."""
        if torch.is_tensor(data):
            return cls(data)
        else:
            return super(TensorColumn, cls).from_data(data)

    def _copy_data(self) -> torch.Tensor:
        return self._data.clone()

    def _view_data(self) -> object:
        return self._data

    def _write_data(self, path: str) -> None:
        # Saving all cell data in a single pickle file
        torch.save(self.data, os.path.join(path, "data.pt"))

    @staticmethod
    def _read_data(path: str) -> torch.Tensor:
        return torch.load(os.path.join(path, "data.pt"))

    def is_equal(self, other: AbstractColumn) -> bool:
        return (other.__class__ == self.__class__) and (self.data == other.data).all()

    def to_tensor(self) -> torch.Tensor:
        return self.data



from __future__ import annotations

import logging
from typing import Sequence

import cytoolz as tz
import numpy as np

from meerkat.cells.abstract import AbstractCell
from meerkat.columns.abstract import AbstractColumn

logger = logging.getLogger(__name__)


class CellColumn(AbstractColumn):
    def __init__(
        self,
        cells: Sequence[AbstractCell] = None,
        *args,
        **kwargs,
    ):
        super(CellColumn, self).__init__(
            data=cells,
            *args,
            **kwargs,
        )

    def _get_cell(self, index: int, materialize: bool = True):
        cell = self._data[index]
        if materialize:
            return cell.get()
        else:
            return cell

    def _get_batch(self, indices: np.ndarray, materialize: bool = True):
        if materialize:
            # if materializing, return a batch (by default, a list of objects returned
            # by `.get`, otherwise the batch format specified by `self.collate`)
            return self.collate([self._data[i].get() for i in indices])

        else:
            return [self._data[i] for i in indices]

    def _get(self, index, materialize: bool = True, _data: np.ndarray = None):
        index = self._translate_index(index)
        if isinstance(index, int):
            if _data is None:
                _data = self._get_cell(index, materialize=materialize)
            return _data

        elif isinstance(index, np.ndarray):
            # support for blocks
            if _data is None:
                _data = self._get_batch(index, materialize=materialize)
            if materialize:
                # materialize could change the data in unknown ways, cannot clone
                return self.__class__.from_data(data=_data)
            else:
                return self._clone(data=_data)

    @classmethod
    def from_cells(cls, cells: Sequence[AbstractCell], *args, **kwargs):
        return cls(cells=cells, *args, **kwargs)

    @property
    def cells(self):
        return self.data

    def _repr_cell(self, index) -> object:
        return self.lz[index].__repr__()

    @staticmethod
    def concat(columns: Sequence[CellColumn]):
        return columns[0].__class__.from_cells(
            list(tz.concat([c.data for c in columns]))
        )

    def is_equal(self, other: AbstractColumn) -> bool:
        return (
            (self.__class__ == other.__class__)
            and (len(self) == len(other))
            and all([self.lz[idx] == other.lz[idx] for idx in range(len(self))])
        )



from __future__ import annotations

import logging
from typing import Sequence

from meerkat.cells.volume import MedicalVolumeCell
from meerkat.columns.cell_column import CellColumn

logger = logging.getLogger(__name__)


class MedicalVolumeColumn(CellColumn):
    def __init__(self, *args, **kwargs):
        super(MedicalVolumeColumn, self).__init__(*args, **kwargs)

    @classmethod
    def from_filepaths(
        cls,
        filepaths: Sequence[str] = None,
        loader: callable = None,
        transform: callable = None,
        *args,
        **kwargs,
    ):
        if loader is None:
            loader = MedicalVolumeCell.default_loader(filepaths)

        cells = [
            MedicalVolumeCell(fp, transform=transform, loader=loader)
            for fp in filepaths
        ]

        return cls(
            cells=cells,
            *args,
            **kwargs,
        )



from functools import reduce
from typing import Union

import torch

from meerkat.ml.model import Model


class ActivationExtractor:
    """Class for extracting activations of a targeted intermediate layer."""

    def __init__(self):
        self.activation = None

    def add_hook(self, module, input, output):
        self.activation = output


class ActivationOp:
    def __init__(
        self,
        model: Union[torch.nn.Module, Model],
        target_module: str,  # TODO(Priya): Support multiple extraction layers
        device: int = None,
    ):
        """Registers a forward hook on the target layer to facilitate
        extracting model activations.

        Args:
            model (nn.Module): the torch model from which activations are extracted
            target_module (str): the name of the submodule of `model` (i.e. an
                intermediate layer) that outputs the activations we'd like to extract.
                For nested submodules, specify a path separated by "." (e.g.
                `ActivationCachedOp(model, "block4.conv")`).
            device (int, optional): the device for the forward pass. Defaults to None,
                in which case the CPU is used.
        """
        self.model = model
        self.device = device
        self.target_module = target_module

        try:
            target_module = _nested_getattr(
                model.model if isinstance(model, Model) else model, target_module
            )
        except AttributeError:
            raise ValueError(f"`model` does not have a submodule {target_module}")

        self.extractor = ActivationExtractor()
        target_module.register_forward_hook(self.extractor.add_hook)

        if not device:
            self.device = "cpu"
            if torch.cuda.is_available():
                self.device = "cuda:0"
        self.model.to(self.device)


def _nested_getattr(obj, attr, *args):
    """Get a nested property from an object.

    Example:
    ```
        model = ...
        weights = _nested_getattr(model, "layer4.weights")
    ```
    """

    return reduce(lambda o, a: getattr(o, a, *args), [obj] + attr.split("."))



from __future__ import annotations

import abc
import logging
import os
from typing import Sequence, Text

import yaml
from yaml.representer import Representer

from meerkat.columns.list_column import ListColumn
from meerkat.tools.lazy_loader import LazyLoader

spacy = LazyLoader("spacy")
spacy_attrs = LazyLoader("spacy.attrs")
spacy_tokens = LazyLoader("spacy.tokens")

Representer.add_representer(abc.ABCMeta, Representer.represent_name)

logger = logging.getLogger(__name__)


class SpacyColumn(ListColumn):
    def __init__(
        self,
        data: Sequence[spacy_tokens.Doc] = None,
        *args,
        **kwargs,
    ):
        super(SpacyColumn, self).__init__(data=data, *args, **kwargs)

    @classmethod
    def from_docs(cls, data: Sequence[spacy_tokens.Doc], *args, **kwargs):
        return cls(data=data, *args, **kwargs)

    @classmethod
    def from_texts(
        cls,
        texts: Sequence[Text],
        lang: str = "en_core_web_sm",
        *args,
        **kwargs,
    ):
        # Create the pipeline
        nlp = spacy.load(lang)
        return cls(data=[nlp(text) for text in texts], *args, **kwargs)

    @property
    def docs(self):
        return self.data

    @property
    def tokens(self):
        return [list(doc) for doc in self]

    def __getattr__(self, item):
        try:
            return [getattr(doc, item) for doc in self]
        except AttributeError:
            raise AttributeError(f"Attribute {item} not found.")

    @classmethod
    def read(
        cls,
        path: str,
        nlp: spacy.language.Language = None,
        lang: str = None,
        *args,
        **kwargs,
    ) -> SpacyColumn:
        assert (nlp is None) != (lang is None)
        if nlp is None:
            nlp = spacy.load(lang)

        # Load in the data
        metadata = dict(
            yaml.load(open(os.path.join(path, "meta.yaml")), Loader=yaml.FullLoader)
        )
        assert metadata["dtype"] == cls

        # Load the `DocBin` from disk
        docbin = spacy_tokens.DocBin().from_disk(os.path.join(path, "data.spacy"))
        return cls(list(docbin.get_docs(nlp.vocab)))

    def write(self, path: str, **kwargs) -> None:
        # Construct the metadata
        state = self._get_state()

        metadata = {
            "dtype": type(self),
            "len": len(self),
            "state": state,
            **self.metadata,
        }

        # Make directory
        os.makedirs(path, exist_ok=True)

        # Get the paths where metadata and data should be stored
        metadata_path = os.path.join(path, "meta.yaml")
        data_path = os.path.join(path, "data.spacy")

        # Create a `DocBin` to store the docs
        attrs = [name for name in spacy_attrs.NAMES if name != "HEAD"]
        docbin = spacy_tokens.DocBin(attrs=attrs, store_user_data=True, docs=self.docs)

        # Save all the docs
        docbin.to_disk(data_path)

        # Save the metadata as a yaml
        yaml.dump(metadata, open(metadata_path, "w"))



"""WILDS configuration defaults and operations.

All default configurations are integrated from the WILDS repository:
https://github.com/p-lambda/wilds/blob/ca7e4fd345aa4f998b8322191b083444d85e2a08/examples/configs/datasets.py
"""

base_config = {
    "split_scheme": None,
    "dataset_kwargs": {},
    "download": False,
    "frac": 1.0,
    "version": None,
    "loader_kwargs": {},
    "train_loader": None,
    "uniform_over_groups": None,
    "distinct_groups": None,
    "n_groups_per_batch": None,
    "batch_size": None,
    "eval_loader": "standard",
    "model": None,
    "model_kwargs": {},
    "train_transform": None,
    "eval_transform": None,
    "target_resolution": None,
    "resize_scale": None,
    "max_token_length": None,
    "loss_function": None,
    "groupby_fields": None,
    "group_dro_step_size": None,
    "coral_penalty_weight": None,
    "irm_lambda": None,
    "irm_penalty_anneal_iters": None,
    "algo_log_metric": None,
    "val_metric": None,
    "val_metric_decreasing": None,
    "n_epochs": None,
    "optimizer": None,
    "lr": None,
    "weight_decay": None,
    "max_grad_norm": None,
    "optimizer_kwargs": {},
    "scheduler": None,
    "scheduler_kwargs": {},
    "scheduler_metric_split": "val",
    "scheduler_metric_name": None,
    "process_outputs_function": None,
    "evaluate_all_splits": True,
    "eval_splits": [],
    "eval_only": False,
    "eval_epoch": None,
    "device": 0,
    "seed": 0,
    "log_dir": "./logs",
    "log_every": 50,
    "save_step": None,
    "save_best": True,
    "save_last": True,
    "no_group_logging": None,
    "use_wandb": False,
    "progress_bar": False,
    "resume": False,
}

dataset_defaults = {
    "amazon": {
        "split_scheme": "official",
        "model": "distilbert-base-uncased",
        "train_transform": "bert",
        "eval_transform": "bert",
        "max_token_length": 512,
        "loss_function": "cross_entropy",
        "algo_log_metric": "accuracy",
        "batch_size": 8,
        "lr": 1e-5,
        "weight_decay": 0.01,
        "n_epochs": 3,
        "n_groups_per_batch": 2,
        "irm_lambda": 1.0,
        "coral_penalty_weight": 1.0,
        "loader_kwargs": {
            "num_workers": 1,
            "pin_memory": True,
        },
        "process_outputs_function": "multiclass_logits_to_pred",
    },
    "bdd100k": {
        "split_scheme": "official",
        "model": "resnet50",
        "model_kwargs": {"pretrained": True},
        "loss_function": "multitask_bce",
        "val_metric": "acc_all",
        "val_metric_decreasing": False,
        "optimizer": "SGD",
        "optimizer_kwargs": {"momentum": 0.9},
        "batch_size": 32,
        "lr": 0.001,
        "weight_decay": 0.0001,
        "n_epochs": 10,
        "algo_log_metric": "multitask_binary_accuracy",
        "train_transform": "image_base",
        "eval_transform": "image_base",
        "process_outputs_function": "binary_logits_to_pred",
    },
    "camelyon17": {
        "split_scheme": "official",
        "model": "densenet121",
        "model_kwargs": {"pretrained": False},
        "train_transform": "image_base",
        "eval_transform": "image_base",
        "target_resolution": (96, 96),
        "loss_function": "cross_entropy",
        "groupby_fields": ["hospital"],
        "val_metric": "acc_avg",
        "val_metric_decreasing": False,
        "optimizer": "SGD",
        "optimizer_kwargs": {"momentum": 0.9},
        "scheduler": None,
        "batch_size": 32,
        "lr": 0.001,
        "weight_decay": 0.01,
        "n_epochs": 5,
        "n_groups_per_batch": 2,
        "irm_lambda": 1.0,
        "coral_penalty_weight": 0.1,
        "algo_log_metric": "accuracy",
        "process_outputs_function": "multiclass_logits_to_pred",
    },
    "celebA": {
        "split_scheme": "official",
        "model": "resnet50",
        "model_kwargs": {"pretrained": True},
        "train_transform": "image_base",
        "eval_transform": "image_base",
        "loss_function": "cross_entropy",
        "groupby_fields": ["male", "y"],
        "val_metric": "acc_wg",
        "val_metric_decreasing": False,
        "optimizer": "SGD",
        "optimizer_kwargs": {"momentum": 0.9},
        "scheduler": None,
        "batch_size": 64,
        "lr": 0.001,
        "weight_decay": 0.0,
        "n_epochs": 200,
        "algo_log_metric": "accuracy",
        "process_outputs_function": "multiclass_logits_to_pred",
    },
    "civilcomments": {
        "split_scheme": "official",
        "model": "distilbert-base-uncased",
        "train_transform": "bert",
        "eval_transform": "bert",
        "loss_function": "cross_entropy",
        "groupby_fields": ["black", "y"],
        "val_metric": "acc_wg",
        "val_metric_decreasing": False,
        "batch_size": 16,
        "lr": 1e-5,
        "weight_decay": 0.01,
        "n_epochs": 5,
        "algo_log_metric": "accuracy",
        "max_token_length": 300,
        "irm_lambda": 1.0,
        "coral_penalty_weight": 10.0,
        "loader_kwargs": {
            "num_workers": 1,
            "pin_memory": True,
        },
        "process_outputs_function": "multiclass_logits_to_pred",
    },
    "fmow": {
        "split_scheme": "official",
        "dataset_kwargs": {
            "oracle_training_set": False,
            "seed": 111,
            "use_ood_val": True,
        },
        "model": "densenet121",
        "model_kwargs": {"pretrained": True},
        "train_transform": "image_base",
        "eval_transform": "image_base",
        "loss_function": "cross_entropy",
        "groupby_fields": [
            "year",
        ],
        "val_metric": "acc_worst_region",
        "val_metric_decreasing": False,
        "optimizer": "Adam",
        "scheduler": "StepLR",
        "scheduler_kwargs": {"gamma": 0.96},
        "batch_size": 64,
        "lr": 0.0001,
        "weight_decay": 0.0,
        "n_epochs": 50,
        "n_groups_per_batch": 8,
        "irm_lambda": 1.0,
        "coral_penalty_weight": 0.1,
        "algo_log_metric": "accuracy",
        "process_outputs_function": "multiclass_logits_to_pred",
    },
    "iwildcam": {
        "loss_function": "cross_entropy",
        "val_metric": "F1-macro_all",
        "model_kwargs": {"pretrained": True},
        "train_transform": "image_base",
        "eval_transform": "image_base",
        "target_resolution": (448, 448),
        "val_metric_decreasing": False,
        "algo_log_metric": "accuracy",
        "model": "resnet50",
        "lr": 3e-5,
        "weight_decay": 0.0,
        "batch_size": 16,
        "n_epochs": 12,
        "optimizer": "Adam",
        "split_scheme": "official",
        "scheduler": None,
        "groupby_fields": [
            "location",
        ],
        "n_groups_per_batch": 2,
        "irm_lambda": 1.0,
        "coral_penalty_weight": 10.0,
        "no_group_logging": True,
        "process_outputs_function": "multiclass_logits_to_pred",
    },
    "ogb-molpcba": {
        "split_scheme": "official",
        "model": "gin-virtual",
        "model_kwargs": {"dropout": 0.5},  # include pretrained
        "loss_function": "multitask_bce",
        "groupby_fields": [
            "scaffold",
        ],
        "val_metric": "ap",
        "val_metric_decreasing": False,
        "optimizer": "Adam",
        "batch_size": 32,
        "lr": 1e-03,
        "weight_decay": 0.0,
        "n_epochs": 100,
        "n_groups_per_batch": 4,
        "irm_lambda": 1.0,
        "coral_penalty_weight": 0.1,
        "no_group_logging": True,
        "process_outputs_function": None,
        "algo_log_metric": "multitask_binary_accuracy",
    },
    "py150": {
        "split_scheme": "official",
        "model": "code-gpt-py",
        "loss_function": "lm_cross_entropy",
        "val_metric": "acc",
        "val_metric_decreasing": False,
        "optimizer": "AdamW",
        "optimizer_kwargs": {"eps": 1e-8},
        "lr": 8e-5,
        "weight_decay": 0.0,
        "n_epochs": 3,
        "batch_size": 6,
        "groupby_fields": [
            "repo",
        ],
        "n_groups_per_batch": 2,
        "irm_lambda": 1.0,
        "coral_penalty_weight": 1.0,
        "no_group_logging": True,
        "algo_log_metric": "multitask_accuracy",
        "process_outputs_function": "multiclass_logits_to_pred",
    },
    "poverty": {
        "split_scheme": "official",
        "dataset_kwargs": {
            "no_nl": False,
            "fold": "A",
            "oracle_training_set": False,
            "use_ood_val": True,
        },
        "model": "resnet18_ms",
        "model_kwargs": {"num_channels": 8},
        "train_transform": "poverty_train",
        "eval_transform": None,
        "loss_function": "mse",
        "groupby_fields": [
            "country",
        ],
        "val_metric": "r_wg",
        "val_metric_decreasing": False,
        "algo_log_metric": "mse",
        "optimizer": "Adam",
        "scheduler": "StepLR",
        "scheduler_kwargs": {"gamma": 0.96},
        "batch_size": 64,
        "lr": 0.001,
        "weight_decay": 0.0,
        "n_epochs": 200,
        "n_groups_per_batch": 8,
        "irm_lambda": 1.0,
        "coral_penalty_weight": 0.1,
        "process_outputs_function": None,
    },
    "waterbirds": {
        "split_scheme": "official",
        "model": "resnet50",
        "train_transform": "image_resize_and_center_crop",
        "eval_transform": "image_resize_and_center_crop",
        "resize_scale": 256.0 / 224.0,
        "model_kwargs": {"pretrained": True},
        "loss_function": "cross_entropy",
        "groupby_fields": ["background", "y"],
        "val_metric": "acc_wg",
        "val_metric_decreasing": False,
        "algo_log_metric": "accuracy",
        "optimizer": "SGD",
        "optimizer_kwargs": {"momentum": 0.9},
        "scheduler": None,
        "batch_size": 128,
        "lr": 1e-5,
        "weight_decay": 1.0,
        "n_epochs": 300,
        "process_outputs_function": "multiclass_logits_to_pred",
    },
    "yelp": {
        "split_scheme": "official",
        "model": "bert-base-uncased",
        "train_transform": "bert",
        "eval_transform": "bert",
        "max_token_length": 512,
        "loss_function": "cross_entropy",
        "algo_log_metric": "accuracy",
        "batch_size": 8,
        "lr": 2e-6,
        "weight_decay": 0.01,
        "n_epochs": 3,
        "n_groups_per_batch": 2,
        "process_outputs_function": "multiclass_logits_to_pred",
    },
    "sqf": {
        "split_scheme": "all_race",
        "model": "logistic_regression",
        "train_transform": None,
        "eval_transform": None,
        "model_kwargs": {"in_features": 104},
        "loss_function": "cross_entropy",
        "groupby_fields": ["y"],
        "val_metric": "precision_at_global_recall_all",
        "val_metric_decreasing": False,
        "algo_log_metric": "accuracy",
        "optimizer": "Adam",
        "optimizer_kwargs": {},
        "scheduler": None,
        "batch_size": 4,
        "lr": 5e-5,
        "weight_decay": 0,
        "n_epochs": 4,
        "process_outputs_function": None,
    },
}

##########################################
#   Split-specific defaults for Amazon   #
##########################################

amazon_split_defaults = {
    "official": {
        "groupby_fields": ["user"],
        "val_metric": "10th_percentile_acc",
        "val_metric_decreasing": False,
        "no_group_logging": True,
    },
    "user": {
        "groupby_fields": ["user"],
        "val_metric": "10th_percentile_acc",
        "val_metric_decreasing": False,
        "no_group_logging": True,
    },
    "time": {
        "groupby_fields": ["year"],
        "val_metric": "acc_avg",
        "val_metric_decreasing": False,
    },
    "time_baseline": {
        "groupby_fields": ["year"],
        "val_metric": "acc_avg",
        "val_metric_decreasing": False,
    },
}

user_baseline_splits = [
    "A1CNQTCRQ35IMM_baseline",
    "A1NE43T0OM6NNX_baseline",
    "A1UH21GLZTYYR5_baseline",
    "A20EEWWSFMZ1PN_baseline",
    "A219Y76LD1VP4N_baseline",
    "A37BRR2L8PX3R2_baseline",
    "A3JVZY05VLMYEM_baseline",
    "A9Q28YTLYREO7_baseline",
    "ASVY5XSYJ1XOE_baseline",
    "AV6QDP8Q0ONK4_baseline",
]
for split in user_baseline_splits:
    amazon_split_defaults[split] = {
        "groupby_fields": ["user"],
        "val_metric": "acc_avg",
        "val_metric_decreasing": False,
    }

category_splits = [
    "arts_crafts_and_sewing_generalization",
    "automotive_generalization",
    "books,movies_and_tv,home_and_kitchen,electronics_generalization",
    "books_generalization",
    "category_subpopulation",
    "cds_and_vinyl_generalization",
    "cell_phones_and_accessories_generalization",
    "clothing_shoes_and_jewelry_generalization",
    "digital_music_generalization",
    "electronics_generalization",
    "grocery_and_gourmet_food_generalization",
    "home_and_kitchen_generalization",
    "industrial_and_scientific_generalization",
    "kindle_store_generalization",
    "luxury_beauty_generalization",
    "movies_and_tv,books,home_and_kitchen_generalization",
    "movies_and_tv,books_generalization",
    "movies_and_tv_generalization",
    "musical_instruments_generalization",
    "office_products_generalization",
    "patio_lawn_and_garden_generalization",
    "pet_supplies_generalization",
    "prime_pantry_generalization",
    "sports_and_outdoors_generalization",
    "tools_and_home_improvement_generalization",
    "toys_and_games_generalization",
    "video_games_generalization",
]
for split in category_splits:
    amazon_split_defaults[split] = {
        "groupby_fields": ["category"],
        "val_metric": "acc_avg",
        "val_metric_decreasing": False,
    }

########################################
#   Split-specific defaults for Yelp   #
########################################

yelp_split_defaults = {
    "official": {
        "groupby_fields": ["year"],
        "val_metric": "acc_avg",
        "val_metric_decreasing": False,
    },
    "user": {
        "groupby_fields": ["user"],
        "val_metric": "10th_percentile_acc",
        "val_metric_decreasing": False,
        "no_group_logging": True,
    },
    "time": {
        "groupby_fields": ["year"],
        "val_metric": "acc_avg",
        "val_metric_decreasing": False,
    },
    "time_baseline": {
        "groupby_fields": ["year"],
        "val_metric": "acc_avg",
        "val_metric_decreasing": False,
    },
}

###############################
#   Split-specific defaults   #
###############################

split_defaults = {
    "amazon": amazon_split_defaults,
    "yelp": yelp_split_defaults,
}

loader_defaults = {
    "loader_kwargs": {
        "num_workers": 4,
        "pin_memory": True,
    },
    "n_groups_per_batch": 4,
}


def populate_defaults(config):
    """Populates hyperparameters with defaults implied by choices of other
    hyperparameters."""
    assert config.dataset is not None, "dataset must be specified"
    # implied defaults from choice of dataset
    config = populate_config(config, dataset_defaults[config.dataset])

    # implied defaults from choice of split
    if (
        config.dataset in split_defaults
        and config.split_scheme in split_defaults[config.dataset]
    ):
        config = populate_config(
            config, split_defaults[config.dataset][config.split_scheme]
        )

    # implied defaults from choice of loader
    config = populate_config(config, loader_defaults)

    # basic checks
    required_fields = [
        "split_scheme",
    ]
    for field in required_fields:
        assert (
            getattr(config, field) is not None
        ), f"Must manually specify {field} for this setup."

    return config


def populate_config(config, template: dict, force_compatibility=False):
    """Populates missing (key, val) pairs in config with (key, val) in template.
    Example usage: populate config with defaults
    Args:
        - config: namespace
        - template: dict
        - force_compatibility: option to raise errors if config.key != template[key]
    """
    if template is None:
        return config

    d_config = vars(config)
    for key, val in template.items():
        if not isinstance(val, dict):  # config[key] expected to be a non-index-able
            if key not in d_config or d_config[key] is None:
                d_config[key] = val
            elif d_config[key] != val and force_compatibility:
                raise ValueError(f"Argument {key} must be set to {val}")

        else:  # config[key] expected to be a kwarg dict
            if key not in d_config:
                d_config[key] = {}
            for kwargs_key, kwargs_val in val.items():
                if kwargs_key not in d_config[key] or d_config[key][kwargs_key] is None:
                    d_config[key][kwargs_key] = kwargs_val
                elif d_config[key][kwargs_key] != kwargs_val and force_compatibility:
                    raise ValueError(
                        f"Argument {key}[{kwargs_key}] must be set to {val}"
                    )
    return config



import os

import dill
import yaml

from meerkat.tools.utils import MeerkatLoader


class ColumnIOMixin:
    def write(self, path: str, *args, **kwargs) -> None:
        assert hasattr(self, "_data"), f"{self.__class__.__name__} requires `self.data`"

        # Make all the directories to the path
        os.makedirs(path, exist_ok=True)

        metadata = self._get_meta()

        # Write the state
        self._write_state(path)
        self._write_data(path, *args, **kwargs)

        # Save the metadata as a yaml file
        metadata_path = os.path.join(path, "meta.yaml")
        yaml.dump(metadata, open(metadata_path, "w"))

        return metadata

    def _get_meta(self):
        return {
            "dtype": type(self),
            "len": len(self),
            **self.metadata,
        }

    def _write_data(self, path):
        data_path = os.path.join(path, "data.dill")
        dill.dump(self.data, open(data_path, "wb"))

    def _write_state(self, path):
        state = self._get_state()
        state_path = os.path.join(path, "state.dill")
        dill.dump(state, open(state_path, "wb"))

    @classmethod
    def read(
        cls, path: str, _data: object = None, _meta: object = None, *args, **kwargs
    ) -> object:
        # Assert that the path exists
        assert os.path.exists(path), f"`path` {path} does not exist."

        # Load in the metadata
        meta = (
            dict(
                yaml.load(
                    open(os.path.join(path, "meta.yaml")),
                    Loader=MeerkatLoader,
                )
            )
            if _meta is None
            else _meta
        )

        col_type = meta["dtype"]
        # Load states
        state = col_type._read_state(path)
        data = col_type._read_data(path, *args, **kwargs) if _data is None else _data

        col = col_type.__new__(col_type)
        col._set_state(state)
        col._set_data(data)

        return col

    @staticmethod
    def _read_state(path: str):
        return dill.load(open(os.path.join(path, "state.dill"), "rb"))

    @staticmethod
    def _read_data(path: str):
        return dill.load(open(os.path.join(path, "data.dill"), "rb"))



from __future__ import annotations

import os
import shutil
from collections import defaultdict
from collections.abc import MutableMapping
from typing import Dict, Mapping, Sequence, Union

import numpy as np
import pandas as pd
import yaml

import meerkat.config
from meerkat.block.abstract import AbstractBlock, BlockIndex
from meerkat.columns.abstract import AbstractColumn

from .ref import BlockRef


class BlockManager(MutableMapping):
    """Manages all blocks in a DataPanel."""

    def __init__(self) -> None:
        self._columns: Dict[str, AbstractColumn] = {}  # ordered as of 3.7
        self._column_to_block_id: Dict[str, int] = {}
        self._block_refs: Dict[int, BlockRef] = {}

    def update(self, block_ref: BlockRef):
        """data (): a single blockable object, potentially contains multiple
        columns."""
        for name in block_ref:
            if name in self:
                self.remove(name)

        # although we can't have the same column living in multiple managers
        # we don't view here because it can lead to multiple calls to clone
        self._columns.update(block_ref)

        block_id = id(block_ref.block)
        # check if there already is a block_ref in the manager for this block
        if block_id in self._block_refs:
            self._block_refs[block_id].update(block_ref)
        else:
            self._block_refs[block_id] = block_ref

        self._column_to_block_id.update({name: block_id for name in block_ref.keys()})

    def apply(self, method_name: str = "_get", *args, **kwargs) -> BlockManager:
        """[summary]

        Args:
            fn (str): a function that is applied to a block and column_spec and
                returns a new block and column_spec.
        Returns:
            [type]: [description]
        """
        results = None
        for block_ref in self._block_refs.values():
            result = block_ref.apply(method_name=method_name, *args, **kwargs)
            if results is None:
                results = BlockManager() if isinstance(result, BlockRef) else {}
            results.update(result)

        # apply method to columns not stored in block
        for name, col in self._columns.items():
            if results is not None and name in results:
                continue

            result = getattr(col, method_name)(*args, **kwargs)

            if results is None:
                results = BlockManager() if isinstance(result, AbstractColumn) else {}

            results[name] = result

        if isinstance(results, BlockManager):
            results.reorder(self.keys())
        return results

    def consolidate(self):
        column_order = list(
            self._columns.keys()
        )  # need to maintain order after consolidate

        block_ref_groups = defaultdict(list)
        for block_ref in self._block_refs.values():
            block_ref_groups[block_ref.block.signature].append(block_ref)

        for block_refs in block_ref_groups.values():
            if len(block_refs) == 1:
                # if there is only one block ref in the group, do not consolidate
                continue

            # consolidate group
            block_class = block_refs[0].block.__class__
            block_ref = block_class.consolidate(block_refs)
            self.update(block_ref)

        self.reorder(column_order)

    def remove(self, name):
        if name not in self._columns:
            raise ValueError(f"Remove failed: no column '{name}' in BlockManager.")

        self._columns.pop(name)

        if name in self._column_to_block_id:
            # column is blockable
            block_ref = self._block_refs[self._column_to_block_id[name]]
            del block_ref[name]

            if len(block_ref) == 0:
                self._block_refs.pop(self._column_to_block_id[name])

            self._column_to_block_id.pop(name)

    def reorder(self, order: Sequence[str]):
        if set(order) != set(self._columns):
            raise ValueError("Must include all columns when reordering a BlockManager.")
        self._columns = {name: self._columns[name] for name in order}

    def __getitem__(
        self, index: Union[str, Sequence[str]]
    ) -> Union[AbstractColumn, BlockManager]:
        if isinstance(index, str):
            return self._columns[index]
        elif isinstance(index, Sequence):
            mgr = BlockManager()
            block_id_to_names = defaultdict(list)
            for name in index:
                if name not in self._column_to_block_id:
                    if name in self:
                        # non-blockable column
                        mgr.add_column(col=self._columns[name], name=name)
                    else:
                        raise ValueError(
                            f"`BlockManager` does not contain column '{name}'."
                        )
                else:
                    # group blockable columns by block
                    block_id_to_names[self._column_to_block_id[name]].append(name)

            # block refs for blockable columns
            for block_id, names in block_id_to_names.items():
                block_ref = self._block_refs[block_id]
                mgr.update(block_ref[names])
            mgr.reorder(order=index)
            return mgr
        else:
            raise ValueError(
                f"Unsupported index of type `{type(index)}` passed to `BlockManager`."
            )

    def __setitem__(self, index: str, data: Union[str, Sequence[str]]):
        if isinstance(data, AbstractColumn):
            self.add_column(data, name=index)
        else:
            raise ValueError(
                f"Cannot set item with object of type `{type(data)}` on `BlockManager`."
            )

    def __delitem__(self, key):
        self.remove(key)

    def __len__(self):
        return len(self._columns)

    @property
    def nrows(self):
        return 0 if len(self) == 0 else len(next(iter(self._columns.values())))

    @property
    def ncols(self):
        return len(self)

    def __contains__(self, value):
        return value in self._columns

    def __iter__(self):
        return iter(self._columns)

    def get_block_ref(self, name: str):
        return self._block_refs[self._column_to_block_id[name]]

    def add_column(self, col: AbstractColumn, name: str):
        """Convert data to a meerkat column using the appropriate Column
        type."""
        if len(self) > 0 and len(col) != self.nrows:
            raise ValueError(
                f"Cannot add column '{name}' with length {len(col)} to `BlockManager` "
                f" with length {self.nrows} columns."
            )

        if not col.is_blockable():
            col = col.view()
            self._columns[name] = col

        else:
            col = col.view()
            self.update(BlockRef(columns={name: col}, block=col._block))

    @classmethod
    def from_dict(cls, data: Mapping[str, object]):
        mgr = cls()
        for name, data in data.items():
            col = AbstractColumn.from_data(data)
            mgr.add_column(col=col, name=name)
        return mgr

    def write(self, path: str):
        meta = {
            "dtype": BlockManager,
            "columns": {},
            "_column_order": list(self.keys()),
        }

        # prepare directories
        columns_dir = os.path.join(path, "columns")
        blocks_dir = os.path.join(path, "blocks")
        meta_path = os.path.join(path, "meta.yaml")
        if os.path.isdir(path):
            if (
                os.path.exists(meta_path)
                and os.path.exists(columns_dir)
                and os.path.exists(blocks_dir)
            ):
                # if overwriting, ensure that old columns are removed
                shutil.rmtree(columns_dir)
                shutil.rmtree(blocks_dir)
            else:
                # if path already points to a dir that wasn't previously holding a
                # block manager, do not overwrite it. We'd like to protect against
                # situation in which user accidentally puts in an important directory
                raise IsADirectoryError(
                    f"Cannot write `BlockManager`. {path} is a directory."
                )

        os.makedirs(path, exist_ok=True)
        os.makedirs(blocks_dir)
        os.makedirs(columns_dir)

        # consolidate before writing
        self.consolidate()
        for block_id, block_ref in self._block_refs.items():
            block: AbstractBlock = block_ref.block
            block_dir = os.path.join(blocks_dir, str(block_id))
            block.write(block_dir)

            for name, column in block_ref.items():
                column_dir = os.path.join(columns_dir, name)
                os.makedirs(column_dir, exist_ok=True)

                # don't write the data, reference the block
                meta["columns"][name] = {
                    **column._get_meta(),
                    "block": {
                        "block_dir": os.path.relpath(block_dir, path),
                        "block_index": _serialize_block_index(column._block_index),
                        "mmap": block.is_mmap,
                    },
                }
                column._write_state(column_dir)

        # write columns not in a block
        for name, column in self._columns.items():
            if name in meta["columns"]:
                continue
            meta["columns"][name] = column._get_meta()
            column.write(os.path.join(columns_dir, name))

        # Save the metadata as a yaml file
        yaml.dump(meta, open(meta_path, "w"))

    @classmethod
    def read(
        cls,
        path: str,
        columns: Sequence[str] = None,
        *args,
        **kwargs,
    ) -> BlockManager:
        """Load a DataPanel stored on disk."""

        # Load the metadata
        meta = dict(
            yaml.load(open(os.path.join(path, "meta.yaml")), Loader=yaml.FullLoader)
        )

        blocks = {}
        mgr = cls()
        for name, col_meta in meta["columns"].items():
            column_dir = os.path.join(path, "columns", name)
            # load a subset of columns
            if columns is not None and name not in columns:
                continue

            if "block" in col_meta:
                # read block or fetch it from `blocks` if it's already been read
                block_meta = col_meta["block"]
                if block_meta["block_dir"] not in blocks:
                    blocks[block_meta["block_dir"]] = AbstractBlock.read(
                        os.path.join(path, block_meta["block_dir"]),
                        mmap=block_meta.get("mmap", False),
                    )
                block = blocks[block_meta["block_dir"]]

                # read column, passing in a block_view
                col = col_meta["dtype"].read(
                    column_dir,
                    _data=block[_deserialize_block_index(block_meta["block_index"])],
                    _meta=col_meta,
                    **kwargs,
                )
                mgr.add_column(col, name)
            else:
                mgr.add_column(
                    col_meta["dtype"].read(path=column_dir, _meta=col_meta, **kwargs),
                    name,
                )
        mgr.reorder(meta["_column_order"])
        return mgr

    def _repr_pandas_(self, max_rows: int = None):
        if max_rows is None:
            max_rows = meerkat.config.DisplayOptions.max_rows
        cols = {}
        formatters = {}
        for name, column in self._columns.items():
            cols[name], formatters[name] = column._repr_pandas_(max_rows=max_rows)
        if self.nrows > max_rows:
            pd_index = np.concatenate(
                (
                    np.arange(max_rows // 2),
                    np.zeros(1),
                    np.arange(self.nrows - max_rows // 2, self.nrows),
                ),
            )
        else:
            pd_index = np.arange(self.nrows)

        df = pd.DataFrame(cols)
        df = df.set_index(pd_index.astype(int))
        return df, formatters

    def view(self):
        mgr = BlockManager()
        for name, col in self.items():
            mgr.add_column(col.view(), name)
        return mgr

    def copy(self):
        mgr = BlockManager()
        for name, col in self.items():
            mgr.add_column(col.copy(), name)
        return mgr


def _serialize_block_index(index: BlockIndex) -> Union[Dict, str, int]:
    if not isinstance(index, (int, str, slice)):
        raise ValueError("Can only serialize `BlockIndex` objects.")
    elif isinstance(index, slice):
        return {"start": index.start, "stop": index.stop, "step": index.step}
    return index


def _deserialize_block_index(index: Union[Dict, int, str]) -> BlockIndex:
    if isinstance(index, Dict):
        return slice(index["start"], index["stop"], index["step"])
    return index



class MergeError(ValueError):
    pass


class ConcatError(ValueError):
    pass


class ConcatWarning(RuntimeWarning):
    pass


class ConsolidationError(ValueError):
    pass


class ImmutableError(ValueError):
    pass


class ExperimentalWarning(FutureWarning):
    pass



from collections.abc import Callable
from typing import List


def identity_collate(batch: List):
    return batch


class CollateMixin:
    def __init__(self, collate_fn: Callable = None, *args, **kwargs):
        super(CollateMixin, self).__init__(*args, **kwargs)

        if collate_fn is not None:
            self._collate_fn = collate_fn
        else:
            self._collate_fn = identity_collate

    @property
    def collate_fn(self):
        """Method used to collate."""
        return self._collate_fn

    @collate_fn.setter
    def collate_fn(self, value):
        self._collate_fn = value

    def collate(self, *args, **kwargs):
        """Collate data."""
        return self._collate_fn(*args, **kwargs)



from collections.abc import Callable, Mapping, Sequence
from types import SimpleNamespace

import numpy as np
import torch


class FunctionInspectorMixin:
    def __init__(self, *args, **kwargs):
        super(FunctionInspectorMixin, self).__init__(*args, **kwargs)

    def _inspect_function(
        self,
        function: Callable,
        with_indices: bool = False,
        is_batched_fn: bool = False,
        data=None,
        indices=None,
        materialize=True,
        **kwargs
    ) -> SimpleNamespace:

        # Initialize variables to track
        no_output = dict_output = bool_output = list_output = False

        # If dict_output = True and `function` is used for updating the `DataPanel`
        # useful to know if any existing column is modified
        updates_existing_column = True
        existing_columns_updated = []

        # Record the shape and dtype of the output
        output_shape = output_dtype = None

        # Run the function to test it
        if data is None:
            if is_batched_fn:
                data = self[:2] if materialize else self.lz[:2]
            else:
                data = self[0] if materialize else self.lz[0]

        if indices is None:
            if is_batched_fn:
                indices = range(2)
            else:
                indices = 0

        if with_indices and is_batched_fn:
            output = function(data, indices, **kwargs)
        elif with_indices and not is_batched_fn:
            output = function(data, indices, **kwargs)
        else:
            output = function(data, **kwargs)

        # lazy import to avoid circular dependency
        from meerkat.columns.abstract import AbstractColumn
        from meerkat.columns.numpy_column import NumpyArrayColumn
        from meerkat.columns.tensor_column import TensorColumn

        if isinstance(output, Mapping):
            # `function` returns a dict output
            dict_output = True

            # Check if `self` is a `DataPanel`
            if hasattr(self, "all_columns"):
                # Set of columns that are updated
                existing_columns_updated = set(self.all_columns).intersection(
                    set(output.keys())
                )

                # Check if `function` updates an existing column
                if len(existing_columns_updated) == 0:
                    updates_existing_column = False

        elif output is None:
            # `function` returns None
            no_output = True

        elif (
            isinstance(output, (bool, np.bool_))
            or (
                isinstance(output, (np.ndarray, NumpyArrayColumn))
                and output.dtype == bool
            )
            or (
                isinstance(output, (torch.Tensor, TensorColumn))
                and output.dtype == torch.bool
            )
        ):

            # `function` returns a bool
            bool_output = True

        elif isinstance(output, (Sequence, AbstractColumn, torch.Tensor, np.ndarray)):
            # `function` returns a list
            list_output = True
            if is_batched_fn and (
                isinstance(output[0], (bool, np.bool_))
                or (isinstance(output[0], np.ndarray) and (output[0].dtype == bool))
                or (
                    isinstance(output[0], torch.Tensor)
                    and (output[0].dtype == torch.bool)
                )
            ):
                # `function` returns a bool per example
                bool_output = True

        if not isinstance(output, Mapping) and not isinstance(output, bool):
            # Record the shape of the output
            if hasattr(output, "shape"):
                output_shape = output.shape
            elif len(output) > 0 and hasattr(output[0], "shape"):
                output_shape = output[0].shape

            # Record the dtype of the output
            if hasattr(output, "dtype"):
                output_dtype = output.dtype
            elif len(output) > 0 and hasattr(output[0], "dtype"):
                output_dtype = output[0].dtype

        return SimpleNamespace(
            output=output,
            dict_output=dict_output,
            no_output=no_output,
            bool_output=bool_output,
            list_output=list_output,
            updates_existing_column=updates_existing_column,
            existing_columns_updated=existing_columns_updated,
            output_shape=output_shape,
            output_dtype=output_dtype,
        )
